# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Implements a class used to store and manipulate covariance matrices.
"""

import warnings

import numpy as np

from astropy import table
from astropy.units import Quantity
from astropy.utils.compat.optional_deps import HAS_SCIPY
from astropy.utils.exceptions import AstropyUserWarning

from .nddata import NDUncertainty

# Required scipy.sparse imports
if HAS_SCIPY:
    from scipy.sparse import coo_matrix, csr_matrix, find, isspmatrix_csr, triu
else:
    err = "Use of 'astropy.nddata.Covariance' requires 'scipy.sparse' module!"

    def find(*args, **kwargs):
        raise ModuleNotFoundError(err)

    def triu(*args, **kwargs):
        raise ModuleNotFoundError(err)

    def csr_matrix(*args, **kwargs):
        raise ModuleNotFoundError(err)

    def coo_matrix(*args, **kwargs):
        raise ModuleNotFoundError(err)

    def isspmatrix_csr(*args, **kwargs):
        raise ModuleNotFoundError(err)


__all__ = ["Covariance"]


# Disabling doctests when scipy isn't present.
__doctest_requires__ = {"*": ["scipy"]}


def _get_csr(arr):
    """
    Helper method used to confirm the array is a `~scipy.sparse.csr_matrix` or
    try to convert it to one.

    Parameters
    ----------
    arr : array-like
        An array that either is a `~scipy.sparse.csr_matrix` or can be converted
        to one.

    Returns
    -------
    `~scipy.sparse.csr_matrix`
        Converted or original matrix.
    """
    # Check the input type
    if isspmatrix_csr(arr):
        return arr
    try:
        return csr_matrix(arr)
    except ValueError:
        raise TypeError(
            "Input matrix is not a scipy.sparse.csr_matrix and could "
            "not be converted to one."
        )


def _impose_sparse_value_threshold(arr, threshold):
    """
    Remove values from a sparse matrix if their absolute value is below the
    provided threshold.

    Parameters
    ----------
    arr : `~scipy.sparse.csr_matrix`
        Array to manipulate.

    threshold : :obj:`float`
        Threshold value

    Returns
    -------
    `~scipy.sparse.csr_matrix`
        Manipulated or original matrix.
    """
    i, j, aij = find(arr)
    index = np.logical_not(np.absolute(aij) < threshold)
    if all(index):
        return arr
    return coo_matrix((aij[index], (i[index], j[index])), shape=arr.shape).tocsr()


def _parse_shape(shape):
    """
    Parse a string representation of an array shape into a tuple.

    Parameters
    ----------
    shape : str
        String representation of the tuple.  It should only contain
        comma-separated integers and parentheses.

    Returns
    -------
    tuple
        Tuple with the shape.
    """
    return tuple([int(n) for n in shape.strip("()").split(",") if len(n) > 0])


class Covariance(NDUncertainty):
    r"""
    A general utility for storing, manipulating, and I/O of covariance matrices.

    Covariance matrices are symmetric by definition, :math:`\Sigma_{ij} =
    \Sigma_{ji}`.  The object therefore only stores the upper triangle of the
    matrix using a `scipy.sparse.csr_matrix`.  By default, instantiation will
    check for symmetry and issue a warning if the matrix is not symmetric.  This
    check can be skipped using the ``assume_symmetric`` keyword.  However, by
    virtue of how the data is stored, symmetry is *always imposed* on the
    matrix.  That is, if a non-symmetric matrix is used to instantiate a
    `Covariance` object, the stored data will yield a matrix that is different
    from the original input.

    Covariance matrices of higher dimensional arrays are always assumed to be
    stored following row-major indexing.  For example, the covariance value
    :math:`\Sigma_{ij}` for an image of size :math:`(N_x,N_y)` is the covariance
    between image pixels :math:`I_{x_i,y_i}` and :math:`I_{x_j,y_j}`, where
    :math:`i = x_i + N_x y_i` and, similarly, :math:`j = x_j + N_x y_j`.

    See :ref:`nddata-covariance` for additional documentation and examples.

    Parameters
    ----------
    array : array-like, `~scipy.sparse.csr_matrix`
        Covariance matrix to store.  If the array is not a
        `~scipy.sparse.csr_matrix` instance, it must be convertible to one.  To
        match the calling sequence for `NDUncertainty`, ``array`` has a default
        value of None, but the array *must* be provided for this `Covariance`
        object.

    data_shape : :obj:`tuple`, optional
        The covariance data is for a higher dimensional array with this shape.
        For example, if the covariance data is for a 2D image with shape
        ``(nx,ny)``, set ``data_shape=(nx,ny)``; the shape of the covariance
        array must then be ``(nx*ny, nx*ny)``.  If None, any higher
        dimensionality is ignored.

    assume_symmetric : bool, optional
        Assume the matrix is symmetric.  This means that a check for symmetry is
        not performed, and the user is not warned if the matrix is not
        symmetric.

    unit : unit-like, optional
        Unit for the covariance values.

    Raises
    ------
    TypeError
        Raised if the input array not a `~scipy.sparse.csr_matrix` object and
        cannot be converted to one.

    ValueError
        Raised if ``data_shape`` is provided and the input covariance matrix
        ``array`` does not have the expected shape or if ``array`` is None.
    """

    def __init__(self, array=None, data_shape=None, assume_symmetric=False, unit=None):
        if array is None:
            raise ValueError("Covariance object cannot be instantiated with None.")

        # Ingest the matrix
        self._cov = triu(
            Covariance._ingest_matrix(array, assume_symmetric=assume_symmetric)
        )
        # Save the diagonal as a variance array for convenience
        self._var = self._cov.diagonal()

        # Set the shape and check it; note self._cov must be defined so that
        # call to self.shape below is valid.
        self._data_shape = data_shape
        if self._data_shape is not None and np.prod(self._data_shape) != self.shape[0]:
            raise ValueError(
                "Product of ``data_shape`` must match the covariance axis length."
            )

        # Workspace for index mapping from flattened to original data arrays
        self._data_index_map = None

        super().__init__(array=self._cov, copy=False, unit=unit)

    @staticmethod
    def _ingest_matrix(arr, assume_symmetric=False):
        """
        Helper method to ingest a covariance or correlation matrix.

        This function converts the input to a '~scipy.sparse.csr_matrix` using
        :func:`_get_csr`, and checks that the array is 2D, square, and symmetric.

        Parameters
        ----------
        arr : array-like
            An array that either is a `~scipy.sparse.csr_matrix` or can be converted
            to one.

        assume_symmetric : bool, optional
            Assume the matrix is symmetric.  This means that a check for
            symmetry is not performed, and the user is not warned if the matrix
            is not symmetric.

        Returns
        -------
        `~scipy.sparse.csr_matrix`
            Converted or original matrix.
        """
        # Make sure it's a sparse matrix or can be converted to one.
        _arr = _get_csr(arr)

        # Check that it's 2D
        if _arr.ndim != 2:
            raise ValueError("Covariance arrays must be 2-dimensional.")
        # Check that it's square
        if _arr.shape[0] != _arr.shape[1]:
            raise ValueError("Covariance matrices must be square.")

        # Skip the symmetry check, if requested
        if assume_symmetric:
            return _arr

        # Check that it's symmetric
        flip_diff = _arr - _arr.T
        if not np.allclose(flip_diff.data, np.zeros_like(flip_diff.data)):
            warnings.warn(
                "Asymmetry detected in covariance/correlation matrix.  Matrix will be modified "
                "to be symmetric using its upper triangle.",
                AstropyUserWarning,
            )
            _arr = triu(_arr) + triu(_arr, 1).T
        return _arr

    @property
    def shape(self):
        """Tuple with the shape of the covariance matrix"""
        return self._cov.shape

    @property
    def nnz(self):
        """
        The number of non-zero (NNZ) elements in the full covariance matrix,
        *including* both the upper and lower triangles.
        """
        return self.stored_nnz * 2 - self._cov.shape[0]

    @property
    def stored_nnz(self):
        """
        The number of non-zero elements stored by the object, which only
        counts the non-zero elements in the upper triangle.
        """
        return self._cov.nnz

    @property
    def variance(self):
        """
        The diagonal of the covariance matrix.
        """
        return self._var

    @variance.setter
    def variance(self, value):
        raise NotImplementedError(
            "Directly setting variance values is not allowed for Covariance objects."
        )

    @property
    def uncertainty_type(self):
        """``"cov"``: `Covariance` implements a covariance matrix."""
        return "cov"

    @property
    def quantity(self):
        """
        The covariance matrix as an dense `~astropy.units.Quantity` object.
        """
        return Quantity(self.to_dense(), self.unit, copy=False, dtype=self._cov.dtype)

    def _data_unit_to_uncertainty_unit(self, value):
        """
        Return the uncertainty unit for covariances given the data unit.
        """
        return value**2

    def __repr__(self):
        return f"<{self.__class__.__name__}; shape = {self.shape}>"

    # Skip error propagation for now
    def _propagate_add(self, other_uncert, result_data, correlation):
        return None

    def _propagate_subtract(self, other_uncert, result_data, correlation):
        return None

    def _propagate_multiply(self, other_uncert, result_data, correlation):
        return None

    def _propagate_divide(self, other_uncert, result_data, correlation):
        return None

    @classmethod
    def from_samples(cls, samples, cov_tol=None, rho_tol=None, **kwargs):
        r"""
        Build a covariance object using discrete samples.

        The covariance is generated using `~numpy.cov` for a set of discretely
        sampled data for an :math:`N`-dimensional parameter space.

        Parameters
        ----------
        samples : `~numpy.ndarray`
            Array with samples drawn from an :math:`N`-dimensional parameter
            space. The shape of the input array must be :math:`N_{\rm par}\times
            N_{\rm samples}`.

        cov_tol : :obj:`float`, optional
            The absolute value of any *covariance matrix* entry less than this
            is assumed to be equivalent to (and set to) 0.

        rho_tol : :obj:`float`, optional
            The absolute value of any *correlation coefficient* less than this
            is assumed to be equivalent to (and set to) 0.

        **kwargs : dict, optional
            Passed directly to main instantiation method.

        Returns
        -------
        `Covariance`
            An :math:`N_{\rm par}\times N_{\rm par}` covariance matrix built
            using the provided samples.

        Raises
        ------
        ValueError
            Raised if the input array is not 2D or if the number of samples (length
            of the second axis) is less than 2.
        """
        if samples.ndim != 2:
            raise ValueError("Input samples for covariance matrix must be a 2D array!")
        if samples.shape[1] < 2:
            raise ValueError("Fewer than two samples provided!")
        return Covariance.from_array(
            np.cov(samples), cov_tol=cov_tol, rho_tol=rho_tol, **kwargs
        )

    @classmethod
    def from_array(cls, covar, cov_tol=None, rho_tol=None, **kwargs):
        r"""
        Define a covariance object from an array.

        .. note::

            The only difference between this method and the direct instantiation
            method (i.e., ``Covariance(array=covar)``) is that it can be used to
            impose tolerances on the covariance value and/or correlation
            coefficients.

        Parameters
        ----------
        covar : array-like
            Array with the covariance data. The object must be either a
            `~scipy.sparse.csr_matrix` or an object that can be converted to
            one.  It must also be 2-dimensional and square.

        cov_tol : :obj:`float`, optional
            The absolute value of any *covariance* matrix entry less than this
            is assumed to be equivalent to (and set to) 0.

        rho_tol : :obj:`float`, optional
            The absolute value of any *correlation coefficient* less than this
            is assumed to be equivalent to (and set to) 0.

        **kwargs : dict, optional
            Passed directly to main instantiation method.

        Returns
        -------
        `Covariance`
            The covariance matrix built using the provided array.
        """
        # Get the assume_symmetric flag, either from kwargs or as the default
        # value
        assume_symmetric = kwargs.pop("assume_symmetric", False)
        # Convert the covariance to a correlation matrix.  If rho_tol is None,
        # this just serves to symmetrize the matrix if it's not already.  Set
        # assume_symmetric to True hereafter
        var, rho = Covariance.to_correlation(covar, assume_symmetric=assume_symmetric)
        if rho_tol is not None:
            rho = _impose_sparse_value_threshold(rho, rho_tol)
        _covar = Covariance.revert_correlation(var, rho, assume_symmetric=True)
        if cov_tol is not None:
            _covar = _impose_sparse_value_threshold(_covar, cov_tol)
        return cls(array=_covar, assume_symmetric=True, **kwargs)

    @classmethod
    def from_table(cls, triu_covar):
        r"""
        Construct the covariance matrix from a table with the non-zero elements
        of the upper triangle of the covariance matrix in coordinate format.

        This is the inverse operation of :func:`to_table`.  The class can read
        covariance data written by other programs *as long as they have a
        commensurate format*; see :func:`to_table`.

        Parameters
        ----------
        triu_covar : `~astropy.table.Table`
            The non-zero elements of the upper triangle of the covariance matrix
            in coordinate format; see :func:`to_table`.

        Returns
        -------
        `Covariance`
            The covariance matrix constructed from the tabulated data.

        Raises
        ------
        ValueError
            Raised if ``triu_covar.meta`` is ``None``, if the provided variance array
            does not have the correct size, or if the data is multidimensional
            and the table columns do not have the right shape.
        """
        # Read shapes
        if "COVSHAPE" not in triu_covar.meta:
            raise ValueError("Table meta dictionary *must* contain COVSHAPE")

        shape = _parse_shape(triu_covar.meta["COVSHAPE"])
        data_shape = (
            _parse_shape(triu_covar.meta["COVDSHP"])
            if "COVDSHP" in triu_covar.meta
            else None
        )

        # Number of non-zero elements
        nnz = len(triu_covar)

        # Read coordinate data
        # WARNING: If the data is written correctly, it should always be true that i<=j
        if data_shape is None:
            i = triu_covar["INDXI"].data
            j = triu_covar["INDXJ"].data
        else:
            ndim = triu_covar["INDXI"].shape[1]
            if len(data_shape) != ndim:
                raise ValueError("Mismatch between COVDSHP keyword and tabulated data.")
            i = np.ravel_multi_index(triu_covar["INDXI"].data.T, data_shape)
            j = np.ravel_multi_index(triu_covar["INDXJ"].data.T, data_shape)

        # Units
        unit = triu_covar.meta.get("BUNIT", None)

        # Set covariance data
        cij = triu_covar["COVARIJ"].data
        # NOTE: the astype conversion of cij when instantiating the matrix below
        # is because of how scipy.sparse restricts instantiation of sparse
        # arrays.  It doesn't like big-endian byte order.  To reproduce the
        # underlying error:
        #   >>> import numpy as np
        #   >>> from scipy.sparse._sputils import getdtype
        #   >>> getdtype(np.dtype('>f8'))
        #   Traceback (most recent call last):
        #   File "<stdin>", line 1, in <module>
        #   File "/Users/westfall/.virtualenvs/astropy/lib/python3.12/site-packages/scipy/sparse/_sputils.py", line 137, in getdtype
        #       raise ValueError(f"scipy.sparse does not support dtype {newdtype.name}. "
        #   ValueError: scipy.sparse does not support dtype float64. The only supported types are: bool, int8, uint8, int16, uint16, int32, uint32, int64, uint64, longlong, ulonglong, float32, float64, longdouble, complex64, complex128, clongdouble.
        #   >>> getdtype(np.dtype('<f8'))
        #   dtype('float64')
        cov = coo_matrix((cij.astype(cij.dtype.type), (i, j)), shape=shape).tocsr()

        # Instantiate.  Set assume_symmetric to true to avoid the warning from
        # the _ingest_matrix method
        return cls(array=cov, data_shape=data_shape, unit=unit, assume_symmetric=True)

    @classmethod
    def from_matrix_multiplication(cls, T, covar, **kwargs):
        r"""
        Construct the covariance matrix that results from a matrix
        multiplication.

        Linear operations on a dataset (e.g., binning or smoothing) can be
        written as matrix multiplications of the form

        .. math::

            {\mathbf y} = {\mathbf T}\ {\mathbf x},

        where :math:`{\mathbf T}` is a transfer matrix of size :math:`N_y\times
        N_x`, :math:`{\mathbf x}` is a vector of size :math:`N_x`, and
        :math:`{\mathbf y}` is a vector of length :math:`{N_y}` that results
        from the multiplication.  If :math:`{\mathbf \Sigma}_x` is the
        covariance matrix for :math:`{\mathbf x}`, then the covariance matrix
        for :math:`{\mathbf Y}` is

        .. math::

            {\mathbf \Sigma}_y = {\mathbf T}\ {\mathbf \Sigma}_x\
            {\mathbf T}^\top.

        If ``covar`` is provided as a vector of length :math:`N_x`, it is
        assumed that the elements of :math:`{\mathbf X}` are independent and the
        provided vector gives the *variance* in each element; i.e., the provided
        data represent the diagonal of :math:`{\mathbf \Sigma}`.

        Parameters
        ----------
        T : `~scipy.sparse.csr_matrix`, `~numpy.ndarray`
            Transfer matrix.  See above.

        covar : `~scipy.sparse.csr_matrix`, `~numpy.ndarray`
            Covariance matrix.  See above.

        **kwargs : dict, optional
            Passed directly to main instantiation method.

        Returns
        -------
        `Covariance`
            The covariance matrix resulting from the matrix multiplication.

        Raises
        ------
        ValueError
            Raised if the provided arrays are not two dimensional or if there is
            a shape mismatch.
        """
        if T.ndim != 2:
            raise ValueError("Input transfer matrix must be two-dimensional.")
        nx = T.shape[1]
        if covar.shape != (nx, nx) and covar.shape != (nx,):
            raise ValueError(
                f"Shape of input variance matrix must be either ({nx}, {nx}) or ({nx},)."
            )
        # If it isn't already, convert T to a csr_matrix
        _T = T if isinstance(T, csr_matrix) else csr_matrix(T)
        # Set the covariance matrix in X
        _covar = (
            coo_matrix((covar, (np.arange(nx), np.arange(nx))), shape=(nx, nx)).tocsr()
            if covar.ndim == 1
            else (covar if isinstance(covar, csr_matrix) else csr_matrix(covar))
        )
        # Construct the covariance matrix
        return cls(_T.dot(_covar.dot(_T.transpose())).tocsr(), **kwargs)

    @classmethod
    def from_variance(cls, variance, **kwargs):
        """
        Construct a diagonal covariance matrix using the provided variance.

        Parameters
        ----------
        variance : `~numpy.ndarray`
            The variance vector.

        **kwargs : dict, optional
            Passed directly to main instantiation method.

        Returns
        -------
        `Covariance`
            The diagonal covariance matrix.
        """
        return cls(csr_matrix(np.diagflat(variance)), **kwargs)

    def to_sparse(self, correlation=False):
        """
        Return the full covariance matrix as a `~scipy.sparse.csr_matrix`
        object.

        This method is essentially equivalent to `to_dense` except that it
        returns a sparse array.

        Parameters
        ----------
        correlation : :obj:`bool`, optional
            Return the *correlation* matrix.  If False, return the covariance
            matrix.

        Returns
        -------
        `~scipy.sparse.csr_matrix`
            The sparse matrix with both the upper and lower triangles filled
            (with symmetric information).
        """
        cov = triu(self._cov) + triu(self._cov, 1).T
        if not correlation:
            return cov
        return Covariance.to_correlation(cov, assume_symmetric=True)[1]

    def apply_new_variance(self, var):
        """
        Using the same correlation coefficients, return a new `Covariance`
        object with the provided variance.

        Parameters
        ----------
        var : array-like
            Variance vector. Must have a length that matches this `Covariance`
            instance; e.g., if this instance is ``cov``, the length of ``var``
            must be ``cov.shape[0]``).  Note that, if the covariance is for
            higher dimensional data, this variance array *must* be flattened to
            1D.

        Returns
        -------
        `Covariance`
            A covariance matrix with the same shape and correlation coefficients
            and this object, but with the provided variance.

        Raises
        ------
        ValueError
            Raised if the length of the variance vector is incorrect.
        """
        _var = np.asarray(var)
        if _var.shape != self._var.shape:
            raise ValueError(
                f"Provided variance has incorrect shape.  Expected {self._var.shape}, "
                f"found {_var.shape}."
            )
        i, j, cij = find(self._cov)
        _cov = coo_matrix(
            (cij * np.sqrt(_var[i] / self._var[i] * _var[j] / self._var[j]), (i, j)),
            shape=self.shape,
        ).tocsr()
        return Covariance(
            array=_cov, data_shape=self._data_shape, assume_symmetric=True
        )

    def copy(self):
        """
        Return a copy of this Covariance object.

        Returns
        -------
        `Covariance`
            A copy of the current covariance matrix.
        """
        # Create the new Covariance instance with a copy of the data
        return Covariance(
            array=self._cov.copy(),
            data_shape=self._data_shape,
            assume_symmetric=True,
            unit=self.unit,
        )

    def to_dense(self, correlation=False):
        """
        Return the full covariance matrix as a `numpy.ndarray` object (a "dense"
        array).

        Parameters
        ----------
        correlation : bool, optional
            Flag to return the correlation matrix, instead of the covariance
            matrix.  Note that setting this to ``True`` does *not* also return the
            variance vector.

        Returns
        -------
        `~numpy.ndarray`
            Dense array with the full covariance matrix.
        """
        return self.to_sparse(correlation=correlation).toarray()

    def find(self, correlation=False):
        """
        Find the non-zero values in the **full** covariance matrix (not just the
        upper triangle).

        This is a simple wrapper for `to_sparse` and `~scipy.sparse.find`.

        Parameters
        ----------
        correlation : bool, optional
            Flag to return the correlation data, instead of the covariance data.
            Note that setting this to ``True`` does *not* also return the variance
            vector.

        Returns
        -------
        i, j : `numpy.ndarray`
            Arrays containing the index coordinates of the non-zero values in
            the covariance (or correlation) matrix.
        c : `numpy.ndarray`
            The non-zero covariance (or correlation) matrix values located at
            the provided ``i,j`` coordinates.
        """
        return find(self.to_sparse(correlation=correlation))

    def covariance_to_data_indices(self, i, j):
        r"""
        Given indices along the two axes of the covariance matrix, return the
        relevant indices in the data array.  This is the inverse of
        :func:`data_to_covariance_indices`.

        Parameters
        ----------
        i : `~numpy.ndarray`
            1D array with the index along the first axis of the covariance
            matrix.  Must be in the range :math:`0...n-1`, where :math:`n` is
            the length of the covariance-matrix axes.

        j : `~numpy.ndarray`
            1D array with the index along the second axis of the covariance
            matrix.  Must be in the range :math:`0...n-1`, where :math:`n` is
            the length of the covariance-matrix axes.

        Returns
        -------
        i_data, i_data : tuple, `numpy.ndarray`
            If `data_shape` is not defined, the input arrays are simply returned
            (and not copied).  Otherwise, the code uses `~numpy.unravel_index`
            to calculate the relevant data-array indices; each element in the
            two-tuple is itself a tuple of :math:`N_{\rm dim}` arrays, one array
            per dimension of the data array.

        Raises
        ------
        ValueError
            Raised if the provided indices fall outside the range of covariance
            matrix.

        """
        if self._data_shape is None:
            if np.any(
                (i < 0) | (i > self.shape[0] - 1) | (j < 0) | (j > self.shape[1] - 1)
            ):
                raise ValueError(
                    "Some indices not valid for covariance matrix with shape "
                    f"{self.shape}."
                )
            return i, j
        return np.unravel_index(
            np.atleast_1d(i).ravel(), self._data_shape
        ), np.unravel_index(np.atleast_1d(j).ravel(), self._data_shape)

    def data_to_covariance_indices(self, i, j):
        r"""
        Given indices of elements in the source data array, return the matrix
        coordinates with the associated covariance.  This is the inverse of
        :func:`covariance_to_data_indices`.

        Parameters
        ----------
        i : array-like, tuple
            A tuple of :math:`N_{\rm dim}` array-like objects providing the
            indices of elements in the N-dimensional data array.  This can be an
            array-like object if ``data_shape`` is undefined, in which case the
            values must be in the range :math:`0...n-1`, where :math:`n` is the
            length of the data array.

        j : array-like, tuple
            The same as ``i``, but providing a second set of coordinates at which
            to access the covariance.

        Returns
        -------
        i_covar, j_covar : `numpy.ndarray`
            Arrays providing the indices in the covariance matrix associated
            with the provided data array coordinates.  If ``data_shape`` is not
            defined, the input arrays are simply returned (and not copied).
            Otherwise, the code uses `~numpy.ravel_multi_index` to calculate the
            relevant covariance indices.

        Raises
        ------
        ValueError
            Raised if the provided indices fall outside the range of data array,
            or if the length of the ``i`` or ``j`` tuples is not :math:`N_{\rm
            dim}`.

        """
        if self._data_shape is None:
            if np.any(
                (i < 0) | (i > self.shape[0] - 1) | (j < 0) | (j > self.shape[1] - 1)
            ):
                raise ValueError(
                    "Some indices not valid for covariance matrix with shape "
                    f"{self.shape}."
                )
            return i, j
        if len(i) != len(self.data_shape):
            raise ValueError(
                "Length of input coordinate list (i) is incorrect; expected "
                f"{len(self.data_shape)}, found {len(i)}"
            )
        if len(j) != len(self.data_shape):
            raise ValueError(
                "Length of input coordinate list (j) is incorrect; expected "
                f"{len(self.data_shape)}, found {len(i)}"
            )
        return np.ravel_multi_index(i, self.data_shape), np.ravel_multi_index(
            j, self.data_shape
        )

    def coordinate_data(self, reshape=False):
        r"""
        Construct data arrays with the non-zero covariance components in
        coordinate format.

        Coordinate format means that the covariance matrix data is provided in
        three columns providing :math:`\Sigma_{ij}` and the (0-indexed) matrix
        coordinates :math:`i,j`.

        This procedure is primarily used when constructing the data arrays for
        storage.  Matching the class convention, the returned data only includes
        the upper triangle.

        Parameters
        ----------
        reshape : :obj:`bool`, optional
            If ``reshape`` is ``True`` and `data_shape` is defined, the :math:`i,j`
            indices are converted to the expected data-array indices; see
            :func:`covariance_to_data_indices`.  These can be reverted to the
            coordinates in the covariance matrix using
            :func:`data_to_covariance_indices`.

        Returns
        -------
        i, j : tuple, `numpy.ndarray`
            The row and column indices, :math:`i,j`: of the covariance matrix.
            If reshaping, these are tuples with the index arrays along each of
            the reshaped axes.
        cij : `numpy.ndarray`
            The covariance, :math:`\Sigma_{ij}`, between array elements at
            indices :math:`i` and :math:`j`.

        Raises
        ------
        ValueError
            Raised if ``reshape`` is True but `data_shape` is undefined.
        """
        if reshape and self._data_shape is None:
            raise ValueError(
                "If reshaping, the shape of the data before flattening to the "
                "covariance array (``data_shape``) must be defined."
            )

        # Get the data (only stores the upper triangle!)
        i, j, cij = find(self._cov)

        # Return the data.
        if reshape:
            # Reshape the indices and the variance array.
            return (
                np.unravel_index(i, self._data_shape),
                np.unravel_index(j, self._data_shape),
                cij,
            )
        return i, j, cij

    def to_table(self):
        r"""
        Return the covariance data in a `~astropy.table.Table` using coordinate
        format.

        Coordinate format means that the covariance matrix data is provided in
        three columns providing :math:`\Sigma_{ij}` and the (0-indexed) matrix
        coordinates :math:`i,j`.

        The output table has three columns:

        - ``'INDXI'``: The row index in the covariance matrix.

        - ``'INDXJ'``: The column index in the covariance matrix.

        - ``'COVARIJ'``: The covariance at the relevant :math:`i,j` coordinate.

        The table also contains the following metadata:

        - ``'COVSHAPE'``: The shape of the covariance matrix

        - ``'BUNIT'``: (If ``unit`` is defined) The string representation of the
          covariance units.

        - ``'COVDSHP'``: (If ``data_shape`` is defined) The shape of the
          associated data array.

        If ``data_shape`` is set, the covariance matrix indices are reformatted
        to match the coordinates in the N-dimensional array.

        .. warning::

            Recall that the storage of covariance matrices for higher
            dimensional data always assumes a row-major storage order.

        Objects instantiated by this method can be used to re-instantiate the
        `Covariance` object using `from_table`.

        Returns
        -------
        `~astropy.table.Table`
            Table with the covoariance matrix in coordinate format and the
            relevant metadata.
        """
        meta = {}
        meta["COVSHAPE"] = str(self.shape)
        if self.unit is not None:
            meta["BUNIT"] = self.unit.to_string()
        reshape = self._data_shape is not None
        i, j, cij = self.coordinate_data(reshape=reshape)
        triu_nnz = cij.size
        if reshape:
            meta["COVDSHP"] = str(self._data_shape)
            i = np.column_stack(i)
            j = np.column_stack(j)
            coo_shape = (i.shape[1],)
        else:
            coo_shape = None
        return table.Table(
            [
                table.Column(
                    data=i, name="INDXI", dtype=int, length=triu_nnz, shape=coo_shape
                ),
                table.Column(
                    data=j, name="INDXJ", dtype=int, length=triu_nnz, shape=coo_shape
                ),
                table.Column(data=cij, name="COVARIJ", dtype=float, length=triu_nnz),
            ],
            meta=meta,
        )

    @property
    def data_shape(self):
        """
        The expected shape of the data array associated with this covariance array.
        """
        return (self.shape[0],) if self._data_shape is None else self._data_shape

    @property
    def data_index_map(self):
        """
        An array mapping the index along each axis of the covariance matrix to
        the shape of the associated data array.
        """
        if self._data_index_map is None:
            self._data_index_map = np.arange(self.shape[0])
            if self._data_shape is not None:
                self._data_index_map = self._data_index_map.reshape(self._data_shape)
        return self._data_index_map

    def match_to_data_slice(self, data_slice):
        """
        Return a new `Covariance` instance that is matched to a slice of its
        parent data array.

        Parameters
        ----------
        data_slice : slice, array-like
            Anything that can be used to slice a `numpy.ndarray`.  To generate a
            slice using syntax that mimics accessing numpy array elements, use
            `numpy.s_`; see examples
            :ref:`here<covariance-match-to-data-slice>`.

        Returns
        -------
        `Covariance`
            A new covariance object for the sliced data array.
        """
        remap = self.data_index_map[data_slice]
        index = remap.ravel()
        return Covariance(
            self.to_sparse()[np.ix_(index, index)],
            data_shape=None if len(remap.shape) == 1 else remap.shape,
        )

    @staticmethod
    def to_correlation(cov, assume_symmetric=False):
        r"""
        Convert a covariance matrix into a correlation matrix by dividing each
        element by the variances.

        Specifically, extract ``var`` (:math:`V_i = C_{ii} \equiv \sigma^2_i`)
        and convert ``cov`` from a covariance matrix with elements
        :math:`C_{ij}` to a correlation matrix with :math:`\rho_{ij}` such that

        .. math::

            C_{ij} \equiv \rho_{ij} \sigma_i \sigma_j.

        To revert a variance vector and correlation matrix back to a covariance
        matrix, use :func:`revert_correlation`.

        Parameters
        ----------
        cov : array-like
            Covariance matrix to convert.  Must be a `~scipy.sparse.csr_matrix`
            instance or convertible to one.

        assume_symmetric : bool, optional
            Assume the matrix is symmetric.  This means that a check for
            symmetry is not performed, and the user is not warned if the matrix
            is not symmetric.

        Returns
        -------
        var : `numpy.ndarray`
            Variance vector
        rho : `~scipy.sparse.csr_matrix`
            Correlation matrix

        Raises
        ------
        ValueError
            Raised if the input array is not 2D and square.
        """
        # Ingest the matrix
        _cov = Covariance._ingest_matrix(cov, assume_symmetric=assume_symmetric)
        # Save the diagonal
        var = _cov.diagonal()
        # Find all the non-zero elements
        i, j, cij = find(_cov)
        rho = coo_matrix(
            (cij / np.sqrt(var[i] * var[j]), (i, j)), shape=_cov.shape
        ).tocsr()
        return var, rho

    @staticmethod
    def revert_correlation(var, rho, assume_symmetric=False):
        r"""
        Revert a variance vector and correlation matrix into a covariance matrix.

        This is the reverse operation of `to_correlation`.

        Parameters
        ----------
        var : `~numpy.ndarray`
            Variance vector.  Length must match the diagonal of ``rho``.

        rho : `~numpy.ndarray`, `~scipy.sparse.csr_matrix`
            Correlation matrix.  Diagonal must have the same length as ``var``.

        assume_symmetric : bool, optional
            Assume the matrix is symmetric.  This means that a check for
            symmetry is not performed, and the user is not warned if the matrix
            is not symmetric.

        Returns
        -------
        `~scipy.sparse.csr_matrix`
            Covariance matrix.
        """
        i, j, rhoij = find(
            Covariance._ingest_matrix(rho, assume_symmetric=assume_symmetric)
        )
        return coo_matrix(
            (rhoij * np.sqrt(var[i] * var[j]), (i, j)), shape=rho.shape
        ).tocsr()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Astronomical and physics constants in SI units.  See :mod:`astropy.constants`
for a complete listing of constants defined in Astropy.
"""

import numpy as np

from .config import codata
from .constant import Constant

# ASTRONOMICAL CONSTANTS


class IAU2015(Constant):
    default_reference = "IAU 2015"
    _registry = {}
    _has_incompatible_units = set()


# DISTANCE

# Astronomical Unit (did not change from 2012)
au = IAU2015(
    "au",
    "Astronomical Unit",
    1.49597870700e11,
    "m",
    0.0,
    "IAU 2012 Resolution B2",
    system="si",
)

# Parsec

pc = IAU2015(
    "pc",
    "Parsec",
    au.value / np.radians(1.0 / 3600.0),
    "m",
    au.uncertainty / np.radians(1.0 / 3600.0),
    "Derived from au + IAU 2015 Resolution B 2 note [4]",
    system="si",
)

# Kiloparsec
kpc = IAU2015(
    "kpc",
    "Kiloparsec",
    1000.0 * au.value / np.radians(1.0 / 3600.0),
    "m",
    1000.0 * au.uncertainty / np.radians(1.0 / 3600.0),
    "Derived from au + IAU 2015 Resolution B 2 note [4]",
    system="si",
)

# Luminosity
L_bol0 = IAU2015(
    "L_bol0",
    "Luminosity for absolute bolometric magnitude 0",
    3.0128e28,
    "W",
    0.0,
    "IAU 2015 Resolution B 2",
    system="si",
)


# SOLAR QUANTITIES

# Solar luminosity
L_sun = IAU2015(
    "L_sun",
    "Nominal solar luminosity",
    3.828e26,
    "W",
    0.0,
    "IAU 2015 Resolution B 3",
    system="si",
)

# Solar mass parameter
GM_sun = IAU2015(
    "GM_sun",
    "Nominal solar mass parameter",
    1.3271244e20,
    "m3 / (s2)",
    0.0,
    "IAU 2015 Resolution B 3",
    system="si",
)

# Solar mass (derived from mass parameter and gravitational constant)
M_sun = IAU2015(
    "M_sun",
    "Solar mass",
    GM_sun.value / codata.G.value,
    "kg",
    ((codata.G.uncertainty / codata.G.value) * (GM_sun.value / codata.G.value)),
    f"IAU 2015 Resolution B 3 + {codata.G.reference}",
    system="si",
)

# Solar radius
R_sun = IAU2015(
    "R_sun",
    "Nominal solar radius",
    6.957e8,
    "m",
    0.0,
    "IAU 2015 Resolution B 3",
    system="si",
)


# OTHER SOLAR SYSTEM QUANTITIES

# Jupiter mass parameter
GM_jup = IAU2015(
    "GM_jup",
    "Nominal Jupiter mass parameter",
    1.2668653e17,
    "m3 / (s2)",
    0.0,
    "IAU 2015 Resolution B 3",
    system="si",
)

# Jupiter mass (derived from mass parameter and gravitational constant)
M_jup = IAU2015(
    "M_jup",
    "Jupiter mass",
    GM_jup.value / codata.G.value,
    "kg",
    ((codata.G.uncertainty / codata.G.value) * (GM_jup.value / codata.G.value)),
    f"IAU 2015 Resolution B 3 + {codata.G.reference}",
    system="si",
)

# Jupiter equatorial radius
R_jup = IAU2015(
    "R_jup",
    "Nominal Jupiter equatorial radius",
    7.1492e7,
    "m",
    0.0,
    "IAU 2015 Resolution B 3",
    system="si",
)

# Earth mass parameter
GM_earth = IAU2015(
    "GM_earth",
    "Nominal Earth mass parameter",
    3.986004e14,
    "m3 / (s2)",
    0.0,
    "IAU 2015 Resolution B 3",
    system="si",
)

# Earth mass (derived from mass parameter and gravitational constant)
M_earth = IAU2015(
    "M_earth",
    "Earth mass",
    GM_earth.value / codata.G.value,
    "kg",
    ((codata.G.uncertainty / codata.G.value) * (GM_earth.value / codata.G.value)),
    f"IAU 2015 Resolution B 3 + {codata.G.reference}",
    system="si",
)

# Earth equatorial radius
R_earth = IAU2015(
    "R_earth",
    "Nominal Earth equatorial radius",
    6.3781e6,
    "m",
    0.0,
    "IAU 2015 Resolution B 3",
    system="si",
)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test numpy functions and ufuncs on Distribution.

The tests here are fairly detailed but do not aim for complete
coverage.  Complete coverage of all numpy functions will be done
in test_function_helpers.

TODO: start test_function_helpers once more functions are supported.
"""

import numpy as np
from numpy.testing import assert_array_equal

from astropy import units as u
from astropy.uncertainty import Distribution


class ArraySetup:
    @classmethod
    def setup_class(cls):
        cls.a = (
            np.array([[[0.0]], [[10.0]]])
            + np.array([[0.0], [1.0], [2.0]])
            + np.arange(4.0) / 10.0
        )
        cls.b = -(np.arange(3.0, 6.0)[:, np.newaxis] + np.arange(4.0) / 10.0)
        cls.da = Distribution(cls.a)
        cls.db = Distribution(cls.b)
        cls.c = np.array([[200.0], [300.0]])


class QuantitySetup(ArraySetup):
    @classmethod
    def setup_class(cls):
        super().setup_class()
        cls.a <<= u.m
        cls.b <<= u.km
        cls.da <<= u.m
        cls.db <<= u.km
        cls.c <<= u.Mm


class TestConcatenation(ArraySetup):
    def test_concatenate(self):
        # Concatenate needs consistent shapes.
        db = self.db[np.newaxis]
        concat_a_b = np.concatenate((self.da, db), axis=0)
        expected_distr = np.concatenate((self.a, self.b[np.newaxis]), axis=0)
        assert_array_equal(concat_a_b.distribution, expected_distr)

    def test_concatenate_not_all_distribution(self):
        concat_c_a = np.concatenate((self.c, self.da), axis=1)
        assert isinstance(concat_c_a, Distribution)
        c_bcst = np.broadcast_to(
            self.c[..., np.newaxis], self.c.shape + (self.da.n_samples,), subok=True
        )

        expected_distr = np.concatenate((c_bcst, self.a), axis=1)
        assert_array_equal(concat_c_a.distribution, expected_distr)


class TestQuantityDistributionConcatenation(TestConcatenation, QuantitySetup):
    pass


class TestBroadcast(ArraySetup):
    def test_broadcast_to(self):
        shape = self.da.shape
        ba = np.broadcast_to(self.db, shape, subok=True)
        assert ba.shape == shape
        expected_distr = np.broadcast_to(self.b, self.a.shape, subok=True)
        assert_array_equal(ba.distribution, expected_distr)

    def test_broadcast_arrays(self):
        bda, bdb, bdc = np.broadcast_arrays(self.da, self.db, self.c, subok=True)
        assert type(bda) is type(bdb) is type(self.da)
        assert type(bdc) is type(self.c)
        ba, bb = np.broadcast_arrays(self.a, self.b, subok=True)
        bc = np.broadcast_to(self.c, self.da.shape, subok=True)
        assert_array_equal(bda.distribution, ba)
        assert_array_equal(bdb.distribution, bb)
        assert_array_equal(bdc, bc)

    def test_broadcast_arrays_subok_false(self):
        # subok affects ndarray subclasses but not distribution itself.
        bda, bdb, bdc = np.broadcast_arrays(self.da, self.db, self.c, subok=False)
        assert type(bda.distribution) is type(bdb.distribution) is np.ndarray
        assert type(bdc) is np.ndarray
        ba, bb = np.broadcast_arrays(self.a, self.b, subok=False)
        bc = np.broadcast_to(self.c, self.da.shape, subok=False)
        assert_array_equal(bda.distribution, ba)
        assert_array_equal(bdb.distribution, bb)
        assert_array_equal(bdc, bc)


class TestQuantityDistributionBroadcast(TestBroadcast, QuantitySetup):
    pass



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Handles the "Console" unit format.
"""

from typing import ClassVar, Literal

from astropy.units.core import UnitBase
from astropy.units.enums import DeprecatedUnitAction

from . import base


class Console(base.Base):
    """
    Output-only format for to display pretty formatting at the
    console.

    For example::

      >>> import astropy.units as u
      >>> print(u.Ry.decompose().to_string('console'))  # doctest: +FLOAT_CMP
      2.1798721*10^-18 m^2 kg s^-2
      >>> print(u.Ry.decompose().to_string('console', fraction='multiline'))  # doctest: +FLOAT_CMP
                       m^2 kg
      2.1798721*10^-18 ------
                        s^2
      >>> print(u.Ry.decompose().to_string('console', fraction='inline'))  # doctest: +FLOAT_CMP
      2.1798721*10^-18 m^2 kg / s^2
    """

    _line: ClassVar[str] = "-"
    _space: ClassVar[str] = " "

    @classmethod
    def _format_superscript(cls, number: str) -> str:
        return f"^{number}"

    @classmethod
    def _format_multiline_fraction(
        cls, scale: str, numerator: str, denominator: str
    ) -> str:
        fraclength = max(len(numerator), len(denominator))
        f = f"{{0:<{len(scale)}s}}{{1:^{fraclength}s}}"

        return "\n".join(
            (
                f.format("", numerator),
                f.format(scale, cls._line * fraclength),
                f.format("", denominator),
            )
        )

    @classmethod
    def to_string(
        cls,
        unit: UnitBase,
        fraction: bool | Literal["inline", "multiline"] = False,
        deprecations: DeprecatedUnitAction = DeprecatedUnitAction.WARN,
    ) -> str:
        # Change default of fraction to False, i.e., we typeset
        # without a fraction by default.
        return super().to_string(unit, fraction=fraction)



# Licensed under a 3-clause BSD style license - see LICENSE.rst


from pathlib import Path

from numpy import get_include as get_numpy_include
from setuptools import Extension

ROOT = Path(__file__).parent.relative_to(Path.cwd())


def get_extensions():
    sources = [ROOT / "_np_utils.pyx", ROOT / "_column_mixins.pyx"]
    include_dirs = [get_numpy_include()]

    exts = [
        Extension(
            name=f"astropy.table.{source.stem}",
            sources=[str(source)],
            include_dirs=include_dirs,
        )
        for source in sources
    ]

    return exts



# Licensed under a 3-clause BSD style license - see LICENSE.rst


from numpy import exp
from numpy.typing import ArrayLike

from astropy.cosmology._src.core import dataclass_decorator
from astropy.cosmology._src.parameter import Parameter
from astropy.cosmology._src.typing import FArray
from astropy.cosmology._src.utils import aszarr, deprecated_keywords
from astropy.units import Quantity

from . import scalar_inv_efuncs
from .base import FLRW, FlatFLRWMixin

__all__ = ["Flatw0wzCDM", "w0wzCDM"]

__doctest_requires__ = {"*": ["scipy"]}


@dataclass_decorator
class w0wzCDM(FLRW):
    """FLRW cosmology with a variable dark energy EoS and curvature.

    The equation for the dark energy equation of state (EoS) uses the simple form:
    :math:`w(z) = w_0 + w_z z`.

    This form is not recommended for z > 1.

    Parameters
    ----------
    H0 : float or scalar quantity-like ['frequency']
        Hubble constant at z = 0. If a float, must be in [km/sec/Mpc].

    Om0 : float
        Omega matter: density of non-relativistic matter in units of the
        critical density at z=0.

    Ode0 : float
        Omega dark energy: density of dark energy in units of the critical
        density at z=0.

    w0 : float, optional
        Dark energy equation of state at z=0. This is pressure/density for
        dark energy in units where c=1.

    wz : float, optional
        Derivative of the dark energy equation of state with respect to z.
        A cosmological constant has w0=-1.0 and wz=0.0.

    Tcmb0 : float or scalar quantity-like ['temperature'], optional
        Temperature of the CMB z=0. If a float, must be in [K]. Default: 0 [K].
        Setting this to zero will turn off both photons and neutrinos
        (even massive ones).

    Neff : float, optional
        Effective number of Neutrino species. Default 3.04.

    m_nu : quantity-like ['energy', 'mass'] or array-like, optional
        Mass of each neutrino species in [eV] (mass-energy equivalency enabled).
        If this is a scalar Quantity, then all neutrino species are assumed to
        have that mass. Otherwise, the mass of each species. The actual number
        of neutrino species (and hence the number of elements of m_nu if it is
        not scalar) must be the floor of Neff. Typically this means you should
        provide three neutrino masses unless you are considering something like
        a sterile neutrino.

    Ob0 : float or None, optional
        Omega baryons: density of baryonic matter in units of the critical
        density at z=0.  If this is set to None (the default), any computation
        that requires its value will raise an exception.

    name : str or None (optional, keyword-only)
        Name for this cosmological object.

    meta : mapping or None (optional, keyword-only)
        Metadata for the cosmology, e.g., a reference.

    Examples
    --------
    >>> from astropy.cosmology import w0wzCDM
    >>> cosmo = w0wzCDM(H0=70, Om0=0.3, Ode0=0.7, w0=-0.9, wz=0.2)

    The comoving distance in Mpc at redshift z:

    >>> z = 0.5
    >>> dc = cosmo.comoving_distance(z)
    """

    w0: Parameter = Parameter(
        default=-1.0, doc="Dark energy equation of state at z=0.", fvalidate="float"
    )
    wz: Parameter = Parameter(
        default=0.0,
        doc="Derivative of the dark energy equation of state w.r.t. z.",
        fvalidate="float",
    )

    def __post_init__(self) -> None:
        super().__post_init__()

        # Please see :ref:`astropy-cosmology-fast-integrals` for discussion
        # about what is being done here.
        if self.Tcmb0.value == 0:
            inv_efunc_scalar = scalar_inv_efuncs.w0wzcdm_inv_efunc_norel
            inv_efunc_scalar_args = (
                self.Om0,
                self.Ode0,
                self.Ok0,
                self.w0,
                self.wz,
            )
        elif not self._nu_info.has_massive_nu:
            inv_efunc_scalar = scalar_inv_efuncs.w0wzcdm_inv_efunc_nomnu
            inv_efunc_scalar_args = (
                self.Om0,
                self.Ode0,
                self.Ok0,
                self.Ogamma0 + self.Onu0,
                self.w0,
                self.wz,
            )
        else:
            inv_efunc_scalar = scalar_inv_efuncs.w0wzcdm_inv_efunc
            inv_efunc_scalar_args = (
                self.Om0,
                self.Ode0,
                self.Ok0,
                self.Ogamma0,
                self._nu_info.neff_per_nu,
                self._nu_info.n_massless_nu,
                self._nu_info.nu_y_list,
                self.w0,
                self.wz,
            )
        object.__setattr__(self, "_inv_efunc_scalar", inv_efunc_scalar)
        object.__setattr__(self, "_inv_efunc_scalar_args", inv_efunc_scalar_args)

    @deprecated_keywords("z", since="7.0")
    def w(self, z: Quantity | ArrayLike) -> FArray:
        r"""Returns dark energy equation of state at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'] or array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        w : ndarray
            The dark energy equation of state.

        Notes
        -----
        The dark energy equation of state is defined as
        :math:`w(z) = P(z)/\rho(z)`, where :math:`P(z)` is the pressure at
        redshift z and :math:`\rho(z)` is the density at redshift z, both in
        units where c=1. Here this is given by :math:`w(z) = w_0 + w_z z`.
        """
        return self.w0 + self.wz * aszarr(z)

    @deprecated_keywords("z", since="7.0")
    def de_density_scale(self, z: Quantity | ArrayLike) -> FArray:
        r"""Evaluates the redshift dependence of the dark energy density.

        Parameters
        ----------
        z : Quantity-like ['redshift'] or array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        I : ndarray
            The scaling of the energy density of dark energy with redshift.

        References
        ----------
        .. [1] Linder, E. (2003). Exploring the Expansion History of the Universe.
               Physics Review Letters, 90(9), 091301.

        Notes
        -----
        The scaling factor, I, is defined by :math:`\rho(z) = \rho_0 I`,
        and in this case is given by ([1]_)

        .. math::

           I = \left(1 + z\right)^{3 \left(1 + w_0 - w_z\right)}
                     \exp \left(3 w_z z\right)
        """
        z = aszarr(z)
        return (z + 1.0) ** (3.0 * (1.0 + self.w0 - self.wz)) * exp(3.0 * self.wz * z)


@dataclass_decorator
class Flatw0wzCDM(FlatFLRWMixin, w0wzCDM):
    """FLRW cosmology with a variable dark energy EoS and no curvature.

    The equation for the dark energy equation of state (EoS) uses the simple form:
    :math:`w(z) = w_0 + w_z z`.

    This form is not recommended for z > 1.

    Parameters
    ----------
    H0 : float or scalar quantity-like ['frequency']
        Hubble constant at z = 0. If a float, must be in [km/sec/Mpc].

    Om0 : float
        Omega matter: density of non-relativistic matter in units of the
        critical density at z=0.

    w0 : float, optional
        Dark energy equation of state at z=0. This is pressure/density for
        dark energy in units where c=1.

    wz : float, optional
        Derivative of the dark energy equation of state with respect to z.
        A cosmological constant has w0=-1.0 and wz=0.0.

    Tcmb0 : float or scalar quantity-like ['temperature'], optional
        Temperature of the CMB z=0. If a float, must be in [K]. Default: 0 [K].
        Setting this to zero will turn off both photons and neutrinos
        (even massive ones).

    Neff : float, optional
        Effective number of Neutrino species. Default 3.04.

    m_nu : quantity-like ['energy', 'mass'] or array-like, optional
        Mass of each neutrino species in [eV] (mass-energy equivalency enabled).
        If this is a scalar Quantity, then all neutrino species are assumed to
        have that mass. Otherwise, the mass of each species. The actual number
        of neutrino species (and hence the number of elements of m_nu if it is
        not scalar) must be the floor of Neff. Typically this means you should
        provide three neutrino masses unless you are considering something like
        a sterile neutrino.

    Ob0 : float or None, optional
        Omega baryons: density of baryonic matter in units of the critical
        density at z=0.  If this is set to None (the default), any computation
        that requires its value will raise an exception.

    name : str or None (optional, keyword-only)
        Name for this cosmological object.

    meta : mapping or None (optional, keyword-only)
        Metadata for the cosmology, e.g., a reference.

    Examples
    --------
    >>> from astropy.cosmology import Flatw0wzCDM
    >>> cosmo = Flatw0wzCDM(H0=70, Om0=0.3, w0=-0.9, wz=0.2)

    The comoving distance in Mpc at redshift z:

    >>> cosmo.comoving_distance(0.5)
    <Quantity 1849.74726272 Mpc>
    """

    def __post_init__(self) -> None:
        super().__post_init__()

        # Please see :ref:`astropy-cosmology-fast-integrals` for discussion
        # about what is being done here.
        if self.Tcmb0.value == 0:
            inv_efunc_scalar = scalar_inv_efuncs.fw0wzcdm_inv_efunc_norel
            inv_efunc_scalar_args = (self.Om0, self.Ode0, self.w0, self.wz)
        elif not self._nu_info.has_massive_nu:
            inv_efunc_scalar = scalar_inv_efuncs.fw0wzcdm_inv_efunc_nomnu
            inv_efunc_scalar_args = (
                self.Om0,
                self.Ode0,
                self.Ogamma0 + self.Onu0,
                self.w0,
                self.wz,
            )
        else:
            inv_efunc_scalar = scalar_inv_efuncs.fw0wzcdm_inv_efunc
            inv_efunc_scalar_args = (
                self.Om0,
                self.Ode0,
                self.Ogamma0,
                self._nu_info.neff_per_nu,
                self._nu_info.n_massless_nu,
                self._nu_info.nu_y_list,
                self.w0,
                self.wz,
            )
        object.__setattr__(self, "_inv_efunc_scalar", inv_efunc_scalar)
        object.__setattr__(self, "_inv_efunc_scalar_args", inv_efunc_scalar_args)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Built-in cosmologies.

See :attr:`~astropy.cosmology.realizations.available` for a full list.
"""

__all__ = [  # noqa: F822, RUF022
    "available",
    "default_cosmology",
    # ----
    "WMAP1",
    "WMAP3",
    "WMAP5",
    "WMAP7",
    "WMAP9",
    "Planck13",
    "Planck15",
    "Planck18",
]

import pathlib
import sys

from astropy.utils.data import get_pkg_data_path

from ._src.core import Cosmology
from ._src.default import default_cosmology

__doctest_requires__ = {"*": ["scipy"]}

_COSMOLOGY_DATA_DIR = pathlib.Path(
    get_pkg_data_path("cosmology", "data", package="astropy")
)
available = (
    "WMAP1",
    "WMAP3",
    "WMAP5",
    "WMAP7",
    "WMAP9",
    "Planck13",
    "Planck15",
    "Planck18",
)


def __getattr__(name: str) -> Cosmology:
    """Make specific realizations from data files with lazy import from ``PEP 562``.

    Raises
    ------
    AttributeError
        If "name" is not in :mod:`astropy.cosmology.realizations`
    """
    if name not in available:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}.")

    cosmo = Cosmology.read(
        str(_COSMOLOGY_DATA_DIR / name) + ".ecsv", format="ascii.ecsv"
    )
    object.__setattr__(
        cosmo,
        "__doc__",
        f"{name} instance of {cosmo.__class__.__qualname__} "
        f"cosmology\n(from {cosmo.meta['reference']})",
    )

    # Cache in this module so `__getattr__` is only called once per `name`.
    setattr(sys.modules[__name__], name, cosmo)

    return cosmo


def __dir__() -> list[str]:
    """Directory, including lazily-imported objects."""
    return __all__



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import inspect
import sys
from contextlib import nullcontext
from io import StringIO

import numpy as np
import pytest

from astropy import units as u
from astropy.cosmology import (
    WMAP1,
    WMAP3,
    WMAP5,
    WMAP7,
    WMAP9,
    CosmologyError,
    FlatLambdaCDM,
    Flatw0waCDM,
    FlatwCDM,
    LambdaCDM,
    Planck13,
    Planck15,
    Planck18,
    w0waCDM,
    w0wzCDM,
    wCDM,
    wpwaCDM,
    z_at_value,
)
from astropy.units import allclose
from astropy.utils.compat.optional_deps import HAS_SCIPY
from astropy.utils.exceptions import AstropyUserWarning


@pytest.mark.skipif(not HAS_SCIPY, reason="test requires scipy")
def test_z_at_value_scalar():
    # These are tests of expected values, and hence have less precision
    # than the roundtrip tests below (test_z_at_value_roundtrip);
    # here we have to worry about the cosmological calculations
    # giving slightly different values on different architectures,
    # there we are checking internal consistency on the same architecture
    # and so can be more demanding
    cosmo = Planck13
    assert allclose(z_at_value(cosmo.age, 2 * u.Gyr), 3.19812268, rtol=1e-6)
    assert allclose(z_at_value(cosmo.lookback_time, 7 * u.Gyr), 0.795198375, rtol=1e-6)
    assert allclose(z_at_value(cosmo.distmod, 46 * u.mag), 1.991389168, rtol=1e-6)
    assert allclose(
        z_at_value(cosmo.luminosity_distance, 1e4 * u.Mpc), 1.36857907, rtol=1e-6
    )
    assert allclose(
        z_at_value(cosmo.luminosity_distance, 26.037193804 * u.Gpc, ztol=1e-10),
        3,
        rtol=1e-9,
    )
    assert allclose(
        z_at_value(cosmo.angular_diameter_distance, 1500 * u.Mpc, zmax=2),
        0.681277696,
        rtol=1e-6,
    )
    assert allclose(
        z_at_value(cosmo.angular_diameter_distance, 1500 * u.Mpc, zmin=2.5),
        3.7914908,
        rtol=1e-6,
    )

    # test behavior when the solution is outside z limits (should
    # raise a CosmologyError)
    with (
        pytest.raises(CosmologyError),
        pytest.warns(AstropyUserWarning, match="fval is not bracketed"),
    ):
        z_at_value(cosmo.angular_diameter_distance, 1500 * u.Mpc, zmax=0.5)

    with (
        pytest.raises(CosmologyError),
        pytest.warns(AstropyUserWarning, match="fval is not bracketed"),
        np.errstate(over="ignore"),
    ):
        z_at_value(cosmo.angular_diameter_distance, 1500 * u.Mpc, zmin=4.0)


@pytest.mark.skipif(not HAS_SCIPY, reason="test requires scipy")
class Test_ZatValue:
    def setup_class(self):
        self.cosmo = Planck13

    def test_broadcast_arguments(self):
        """Test broadcast of arguments."""
        # broadcasting main argument
        assert allclose(
            z_at_value(self.cosmo.age, [2, 7] * u.Gyr),
            [3.1981206134773115, 0.7562044333305182],
            rtol=1e-6,
        )

        # basic broadcast of secondary arguments
        assert allclose(
            z_at_value(
                self.cosmo.angular_diameter_distance,
                1500 * u.Mpc,
                zmin=[0, 2.5],
                zmax=[2, 4],
            ),
            [0.681277696, 3.7914908],
            rtol=1e-6,
        )

        # more interesting broadcast
        assert allclose(
            z_at_value(
                self.cosmo.angular_diameter_distance,
                1500 * u.Mpc,
                zmin=[[0, 2.5]],
                zmax=[2, 4],
            ),
            [[0.681277696, 3.7914908]],
            rtol=1e-6,
        )

    def test_broadcast_bracket(self):
        """`bracket` has special requirements."""
        # start with an easy one
        assert allclose(
            z_at_value(self.cosmo.age, 2 * u.Gyr, bracket=None),
            3.1981206134773115,
            rtol=1e-6,
        )

        # now actually have a bracket
        assert allclose(
            z_at_value(self.cosmo.age, 2 * u.Gyr, bracket=[0, 4]),
            3.1981206134773115,
            rtol=1e-6,
        )

        # now a bad length
        with pytest.raises(ValueError, match="sequence"):
            z_at_value(self.cosmo.age, 2 * u.Gyr, bracket=[0, 4, 4, 5])

        # now the wrong dtype : an ndarray, but not an object array
        with pytest.raises(TypeError, match="dtype"):
            z_at_value(self.cosmo.age, 2 * u.Gyr, bracket=np.array([0, 4]))

        # now an object array of brackets
        bracket = np.array([[0, 4], [0, 3, 4]], dtype=object)
        assert allclose(
            z_at_value(self.cosmo.age, 2 * u.Gyr, bracket=bracket),
            [3.1981206134773115, 3.1981206134773115],
            rtol=1e-6,
        )

    def test_bad_broadcast(self):
        """Shapes mismatch as expected"""
        with pytest.raises(ValueError, match="broadcast"):
            z_at_value(
                self.cosmo.angular_diameter_distance,
                1500 * u.Mpc,
                zmin=[0, 2.5, 0.1],
                zmax=[2, 4],
            )

    def test_scalar_input_to_output(self):
        """Test scalar input returns a scalar."""
        z = z_at_value(
            self.cosmo.angular_diameter_distance, 1500 * u.Mpc, zmin=0, zmax=2
        )
        assert isinstance(z, u.Quantity)
        assert z.dtype == np.float64
        assert z.shape == ()


@pytest.mark.skipif(not HAS_SCIPY, reason="test requires scipy")
def test_z_at_value_verbose(monkeypatch):
    cosmo = Planck13

    # Test the "verbose" flag. Since this uses "print", need to mod stdout
    mock_stdout = StringIO()
    monkeypatch.setattr(sys, "stdout", mock_stdout)

    resx = z_at_value(cosmo.age, 2 * u.Gyr, verbose=True)
    assert str(resx.value) in mock_stdout.getvalue()  # test "verbose" prints res


@pytest.mark.skipif(not HAS_SCIPY, reason="test requires scipy")
@pytest.mark.parametrize("method", ["Brent", "Golden", "Bounded"])
def test_z_at_value_bracketed(method):
    """
    Test 2 solutions for angular diameter distance by not constraining zmin, zmax,
    but setting `bracket` on the appropriate side of the turning point z.
    Setting zmin / zmax should override `bracket`.
    """
    cosmo = Planck13

    if method == "Bounded":
        with pytest.warns(AstropyUserWarning, match="fval is not bracketed"):
            z = z_at_value(cosmo.angular_diameter_distance, 1500 * u.Mpc, method=method)
        if z > 1.6:
            z = 3.7914908
            bracket = (0.9, 1.5)
        else:
            z = 0.6812777
            bracket = (1.6, 2.0)
        with (
            pytest.warns(UserWarning, match="Option 'bracket' is ignored"),
            pytest.warns(AstropyUserWarning, match="fval is not bracketed"),
        ):
            assert allclose(
                z_at_value(
                    cosmo.angular_diameter_distance,
                    1500 * u.Mpc,
                    method=method,
                    bracket=bracket,
                ),
                z,
                rtol=1e-6,
            )
    else:
        assert allclose(
            z_at_value(
                cosmo.angular_diameter_distance,
                1500 * u.Mpc,
                method=method,
                bracket=(0.3, 1.0),
            ),
            0.6812777,
            rtol=1e-6,
        )
        assert allclose(
            z_at_value(
                cosmo.angular_diameter_distance,
                1500 * u.Mpc,
                method=method,
                bracket=(2.0, 4.0),
            ),
            3.7914908,
            rtol=1e-6,
        )
        assert allclose(
            z_at_value(
                cosmo.angular_diameter_distance,
                1500 * u.Mpc,
                method=method,
                bracket=(0.1, 1.5),
            ),
            0.6812777,
            rtol=1e-6,
        )
        assert allclose(
            z_at_value(
                cosmo.angular_diameter_distance,
                1500 * u.Mpc,
                method=method,
                bracket=(0.1, 1.0, 2.0),
            ),
            0.6812777,
            rtol=1e-6,
        )
        with pytest.warns(AstropyUserWarning, match=r"fval is not bracketed"):
            assert allclose(
                z_at_value(
                    cosmo.angular_diameter_distance,
                    1500 * u.Mpc,
                    method=method,
                    bracket=(0.9, 1.5),
                ),
                0.6812777,
                rtol=1e-6,
            )
            assert allclose(
                z_at_value(
                    cosmo.angular_diameter_distance,
                    1500 * u.Mpc,
                    method=method,
                    bracket=(1.6, 2.0),
                ),
                3.7914908,
                rtol=1e-6,
            )
        assert allclose(
            z_at_value(
                cosmo.angular_diameter_distance,
                1500 * u.Mpc,
                method=method,
                bracket=(1.6, 2.0),
                zmax=1.6,
            ),
            0.6812777,
            rtol=1e-6,
        )
        assert allclose(
            z_at_value(
                cosmo.angular_diameter_distance,
                1500 * u.Mpc,
                method=method,
                bracket=(0.9, 1.5),
                zmin=1.5,
            ),
            3.7914908,
            rtol=1e-6,
        )

    if method == "Bounded":
        ctx_bracket = pytest.warns(
            UserWarning, match="Option 'bracket' is ignored by method Bounded"
        )
    else:
        ctx_bracket = nullcontext()

    with (
        pytest.raises(CosmologyError),
        pytest.warns(AstropyUserWarning, match="fval is not bracketed"),
        ctx_bracket,
    ):
        z_at_value(
            cosmo.angular_diameter_distance,
            1500 * u.Mpc,
            method=method,
            bracket=(3.9, 5.0),
            zmin=4.0,
        )


@pytest.mark.skipif(not HAS_SCIPY, reason="test requires scipy")
@pytest.mark.parametrize("method", ["Brent", "Golden", "Bounded"])
def test_z_at_value_unconverged(method):
    """
    Test warnings on non-converged solution when setting `maxfun` to too small iteration number -
    only 'Bounded' returns status value and specific message.
    """
    cosmo = Planck18
    ztol = {"Brent": [1e-4, 1e-4], "Golden": [1e-3, 1e-2], "Bounded": [1e-3, 1e-1]}

    if method == "Bounded":
        ctx = pytest.warns(
            AstropyUserWarning,
            match="Solver returned 1: Maximum number of function calls reached",
        )
    else:
        ctx = pytest.warns(AstropyUserWarning, match="Solver returned None")

    with ctx:
        z0 = z_at_value(
            cosmo.angular_diameter_distance, 1 * u.Gpc, zmax=2, maxfun=13, method=method
        )
    with ctx:
        z1 = z_at_value(
            cosmo.angular_diameter_distance, 1 * u.Gpc, zmin=2, maxfun=13, method=method
        )

    assert allclose(z0, 0.32442, rtol=ztol[method][0])
    assert allclose(z1, 8.18551, rtol=ztol[method][1])


@pytest.mark.skipif(not HAS_SCIPY, reason="test requires scipy")
@pytest.mark.parametrize(
    "cosmo",
    [
        Planck13,
        Planck15,
        Planck18,
        WMAP1,
        WMAP3,
        WMAP5,
        WMAP7,
        WMAP9,
        LambdaCDM,
        FlatLambdaCDM,
        wpwaCDM,
        w0wzCDM,
        wCDM,
        FlatwCDM,
        w0waCDM,
        Flatw0waCDM,
    ],
)
def test_z_at_value_roundtrip(cosmo):
    """
    Calculate values from a known redshift, and then check that
    z_at_value returns the right answer.
    """
    z = 0.5

    # Skip Ok, w, de_density_scale because in the Planck cosmologies
    # they are redshift independent and hence uninvertable,
    # *_distance_z1z2 methods take multiple arguments, so require
    # special handling
    # clone is not a redshift-dependent method
    # nu_relative_density is not redshift-dependent in the WMAP cosmologies
    skip = (
        "Ok",
        "Otot",
        "angular_diameter_distance_z1z2",
        "clone",
        "is_equivalent",
        "de_density_scale",
        "w",
    )
    if str(cosmo.name).startswith("WMAP"):
        skip += ("nu_relative_density",)

    methods = inspect.getmembers(cosmo, predicate=inspect.ismethod)

    for name, func in methods:
        if name.startswith("_") or name in skip:
            continue
        fval = func(z)
        # we need zmax here to pick the right solution for
        # angular_diameter_distance and related methods.
        # Be slightly more generous with rtol than the default 1e-8
        # used in z_at_value
        got = z_at_value(func, fval, bracket=[0.3, 1.0], ztol=1e-12)
        assert allclose(got, z, rtol=2e-11), f"Round-trip testing {name} failed"

    # Test distance functions between two redshifts; only for realizations
    if isinstance(getattr(cosmo, "name", None), str):
        z2 = 2.0
        func_z1z2 = [
            lambda z1: cosmo._comoving_distance_z1z2(z1, z2),
            lambda z1: cosmo._comoving_transverse_distance_z1z2(z1, z2),
            lambda z1: cosmo.angular_diameter_distance_z1z2(z1, z2),
        ]
        for func in func_z1z2:
            fval = func(z)
            assert allclose(z, z_at_value(func, fval, zmax=1.5, ztol=1e-12), rtol=2e-11)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""An extensible ASCII table reader and writer."""

from astropy import config as _config

from . import connect
from .basic import (
    Basic,
    BasicData,
    BasicHeader,
    CommentedHeader,
    Csv,
    NoHeader,
    Rdb,
    Tab,
)
from .cds import Cds
from .core import (
    AllType,
    BaseData,
    BaseHeader,
    BaseInputter,
    BaseOutputter,
    BaseReader,
    BaseSplitter,
    Column,
    ContinuationLinesInputter,
    DefaultSplitter,
    FloatType,
    InconsistentTableError,
    IntType,
    NoType,
    NumType,
    ParameterError,
    StrType,
    TableOutputter,
    WhitespaceSplitter,
    convert_numpy,
    masked,
)
from .daophot import Daophot
from .ecsv import Ecsv
from .fastbasic import (
    FastBasic,
    FastCommentedHeader,
    FastCsv,
    FastNoHeader,
    FastRdb,
    FastTab,
)
from .fixedwidth import (
    FixedWidth,
    FixedWidthData,
    FixedWidthHeader,
    FixedWidthNoHeader,
    FixedWidthSplitter,
    FixedWidthTwoLine,
)
from .html import HTML
from .ipac import Ipac
from .latex import AASTex, Latex, latexdicts
from .mrt import Mrt
from .qdp import QDP
from .rst import RST
from .sextractor import SExtractor
from .tdat import Tdat
from .ui import get_read_trace, get_reader, get_writer, read, set_guess, write


class Conf(_config.ConfigNamespace):
    """
    Configuration parameters for `astropy.io.ascii`.
    """

    guess_limit_lines = _config.ConfigItem(
        10000,
        "When guessing the format of a table, this is the number of lines that "
        "will be used for initial guessing. If the reading succeeds based on this "
        "number of lines, then reading the full table will be attempted. If the reading "
        "based on the subset of lines fails, the format will no longer be considered. "
        "This can be set to `None` to disable the limit",
    )


conf = Conf()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np

from astropy.io.fits.hdu.base import BITPIX2DTYPE
from astropy.io.fits.hdu.compressed._tiled_compression import (
    decompress_image_data_section,
)
from astropy.io.fits.hdu.compressed.utils import _n_tiles
from astropy.utils.shapes import simplify_basic_index

__all__ = ["CompImageSection"]


class CompImageSection:
    """
    Class enabling subsets of CompImageHDU data to be loaded lazily via slicing.

    Slices of this object load the corresponding section of an image array from
    the underlying FITS file, and applies any BSCALE/BZERO factors.

    Section slices cannot be assigned to, and modifications to a section are
    not saved back to the underlying file.

    See the :ref:`astropy:data-sections` section of the Astropy documentation
    for more details.
    """

    def __init__(self, hdu):
        self.hdu = hdu

        self._data_shape = self.hdu.shape
        self._tile_shape = self.hdu.tile_shape

        self._n_dim = len(self._data_shape)
        self._n_tiles = np.array(
            _n_tiles(self._data_shape, self._tile_shape), dtype=int
        )

    @property
    def shape(self):
        return tuple(self._data_shape)

    @property
    def ndim(self):
        return len(self.hdu.shape)

    @property
    def dtype(self):
        return np.dtype(BITPIX2DTYPE[self.hdu._bitpix])

    def __getitem__(self, index):
        if self.hdu._bintable is None:
            return self.hdu.data[index]

        # Shortcut if the whole data is requested (this is used by the
        # data property, so we optimize it as it is frequently used)
        if index is Ellipsis:
            first_tile_index = np.zeros(self._n_dim, dtype=int)
            last_tile_index = self._n_tiles - 1
            data = decompress_image_data_section(
                self.hdu._bintable.data,
                self.hdu.compression_type,
                self.hdu._bintable.header,
                self.hdu._bintable,
                self.hdu.header,
                first_tile_index,
                last_tile_index,
            )
            if self.hdu._do_not_scale_image_data:
                return data
            else:
                return self.hdu._scale_data(data)

        index = simplify_basic_index(index, shape=self._data_shape)

        # Determine for each dimension the first and last tile to extract

        first_tile_index = np.zeros(self._n_dim, dtype=int)
        last_tile_index = np.zeros(self._n_dim, dtype=int)

        final_array_index = []

        for dim, idx in enumerate(index):
            if isinstance(idx, slice):
                if idx.step > 0:
                    first_tile_index[dim] = idx.start // self._tile_shape[dim]
                    last_tile_index[dim] = (idx.stop - 1) // self._tile_shape[dim]
                else:
                    stop = 0 if idx.stop is None else max(idx.stop - 1, 0)
                    first_tile_index[dim] = stop // self._tile_shape[dim]
                    last_tile_index[dim] = idx.start // self._tile_shape[dim]

                # Because slices such as slice(5, 0, 1) can exist (which
                # would be empty) we need to make sure last_tile_index is
                # always larger than first_tile_index
                last_tile_index = np.maximum(last_tile_index, first_tile_index)

                if idx.step < 0 and idx.stop is None:
                    final_array_index.append(idx)
                else:
                    final_array_index.append(
                        slice(
                            idx.start - self._tile_shape[dim] * first_tile_index[dim],
                            idx.stop - self._tile_shape[dim] * first_tile_index[dim],
                            idx.step,
                        )
                    )
            else:
                first_tile_index[dim] = idx // self._tile_shape[dim]
                last_tile_index[dim] = first_tile_index[dim]
                final_array_index.append(
                    idx - self._tile_shape[dim] * first_tile_index[dim]
                )

        data = decompress_image_data_section(
            self.hdu._bintable.data,
            self.hdu.compression_type,
            self.hdu._bintable.header,
            self.hdu._bintable,
            self.hdu.header,
            first_tile_index,
            last_tile_index,
        )

        data = data[tuple(final_array_index)]

        if self.hdu._do_not_scale_image_data:
            return data
        else:
            return self.hdu._scale_data(data)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from astropy import config as _config


class Conf(_config.ConfigNamespace):
    """
    Configuration parameters for `astropy.time`.
    """

    use_fast_parser = _config.ConfigItem(
        ["True", "False", "force"],
        "Use fast C parser for supported time strings formats, including ISO, "
        "ISOT, and YearDayTime. Allowed values are the 'False' (use Python parser),"
        "'True' (use C parser and fall through to Python parser if fails), and "
        "'force' (use C parser and raise exception if it fails). Note that the"
        "options are all strings.",
    )

    masked_array_type = _config.ConfigItem(
        ["astropy", "numpy"],
        'The type of masked array used for masked output data.  Can be "astropy" '
        'for `astropy.utils.masked.Masked` or "numpy" to use `numpy.ma.MaskedArray`. '
        "Note that if `astropy.units.Quantity` is produced, the output always "
        "uses `astropy.utils.masked.Masked`, since `numpy.ma.MaskedArray` does not "
        "work with quantities.",
    )
    # Create a dict of available masked classes for speed.
    # Use local imports so we do not pollute the module namespace.
    from numpy.ma import MaskedArray

    from astropy.utils.masked import Masked

    _MASKED_CLASSES = {"astropy": Masked, "numpy": MaskedArray}

    @property
    def _masked_cls(self):
        """The masked class set by ``masked_array_type``.

        This is |Masked| for "astropy", `numpy.ma.MaskedArray` for "numpy".
        """
        return self._MASKED_CLASSES[self.masked_array_type]


conf = Conf()

# isort: off
from .formats import *
from .core import *

# isort: on



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_array_equal

from astropy import units as u
from astropy.coordinates import Angle, is_rotation_or_reflection, rotation_matrix
from astropy.coordinates.matrix_utilities import (
    angle_axis,
    is_O3,
    is_rotation,
)
from astropy.utils.exceptions import AstropyDeprecationWarning

ROTATION_MATRIX = rotation_matrix(35 * u.deg, "x")
IMPROPER_ROTATION_MATRIX = ROTATION_MATRIX.copy()
IMPROPER_ROTATION_MATRIX[0, 0] *= -1
NON_ORTHOGONAL_MATRIX = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])


def test_rotation_matrix():
    assert_allclose(rotation_matrix(0 * u.deg, "x"), np.eye(3))

    assert_allclose(
        rotation_matrix(90 * u.deg, "y"), [[0, 0, -1], [0, 1, 0], [1, 0, 0]], atol=1e-12
    )

    assert_allclose(
        rotation_matrix(-90 * u.deg, "z"),
        [[0, -1, 0], [1, 0, 0], [0, 0, 1]],
        atol=1e-12,
    )

    assert_allclose(
        rotation_matrix(45 * u.deg, "x"), rotation_matrix(45 * u.deg, [1, 0, 0])
    )
    assert_allclose(
        rotation_matrix(125 * u.deg, "y"), rotation_matrix(125 * u.deg, [0, 1, 0])
    )
    assert_allclose(
        rotation_matrix(-30 * u.deg, "z"), rotation_matrix(-30 * u.deg, [0, 0, 1])
    )

    assert_allclose(
        np.dot(rotation_matrix(180 * u.deg, [1, 1, 0]), [1, 0, 0]),
        [0, 1, 0],
        atol=1e-12,
    )

    # make sure it also works for very small angles
    assert_allclose(
        rotation_matrix(0.000001 * u.deg, "x"),
        rotation_matrix(0.000001 * u.deg, [1, 0, 0]),
    )


@pytest.mark.parametrize(
    "angle", [Angle(0 * u.deg), 0 * u.deg, np.array(0), 0, "0 deg"], ids=type
)
def test_rotation_angle_input_types(angle):
    assert_array_equal(rotation_matrix(angle), np.eye(3), strict=True)


def test_angle_axis():
    m1 = rotation_matrix(35 * u.deg, "x")
    with pytest.warns(AstropyDeprecationWarning):
        an1, ax1 = angle_axis(m1)

    assert an1 - 35 * u.deg < 1e-10 * u.deg
    assert_allclose(ax1, [1, 0, 0])

    m2 = rotation_matrix(-89 * u.deg, [1, 1, 0])
    with pytest.warns(AstropyDeprecationWarning):
        an2, ax2 = angle_axis(m2)

    assert an2 - 89 * u.deg < 1e-10 * u.deg
    assert_allclose(ax2, [-(2**-0.5), -(2**-0.5), 0])


@pytest.mark.parametrize(
    "matrix,expectation",
    [
        pytest.param(ROTATION_MATRIX, np.True_, id="rotation"),
        pytest.param(IMPROPER_ROTATION_MATRIX, np.True_, id="improper rotation"),
        pytest.param(
            np.stack((ROTATION_MATRIX, IMPROPER_ROTATION_MATRIX)),
            (np.True_, np.True_),
            id="proper and improper rotation",
        ),
        pytest.param(NON_ORTHOGONAL_MATRIX, np.False_, id="non-orthogonal matrix"),
        pytest.param(
            np.stack((ROTATION_MATRIX, NON_ORTHOGONAL_MATRIX)),
            (np.True_, np.False_),
            id="rotation and non-orthogonal matrix",
        ),
    ],
)
def test_is_rotation_or_reflection(matrix, expectation):
    assert_array_equal(is_rotation_or_reflection(matrix), expectation, strict=True)


@pytest.mark.parametrize(
    "atol,expectation",
    [
        pytest.param(None, (np.True_, np.False_), id="default atol"),
        pytest.param(1, (np.True_, np.True_), id="very large atol"),
    ],
)
def test_is_rotation_or_reflection_atol(atol, expectation):
    assert_array_equal(
        is_rotation_or_reflection(
            np.stack([ROTATION_MATRIX, 0.5 * ROTATION_MATRIX]), atol
        ),
        expectation,
        strict=True,
    )


def test_is_O3():
    with pytest.warns(
        AstropyDeprecationWarning, match=r"Use is_rotation_or_reflection instead\.$"
    ):
        is_O3(ROTATION_MATRIX)


def test_is_rotation():
    """Test the rotation matrix checker ``is_rotation``."""
    # Normal rotation matrix
    m1 = rotation_matrix(35 * u.deg, "x")
    with pytest.warns(AstropyDeprecationWarning):
        assert is_rotation(m1)
    with pytest.warns(AstropyDeprecationWarning):
        assert is_rotation(m1, allow_improper=True)  # (a less restrictive test)
    # and (M, 3, 3)
    n1 = np.tile(m1, (2, 1, 1))
    with pytest.warns(AstropyDeprecationWarning):
        assert tuple(is_rotation(n1)) == (True, True)  # (show the broadcasting)
    # Test atol parameter
    nn1 = np.tile(0.5 * m1, (2, 1, 1))
    with pytest.warns(AstropyDeprecationWarning):
        assert tuple(is_rotation(nn1)) == (False, False)  # (show the broadcasting)
    with pytest.warns(AstropyDeprecationWarning):
        assert tuple(is_rotation(nn1, atol=10)) == (True, True)

    # Improper rotation (unit rotation + reflection)
    m2 = np.identity(3)
    m2[0, 0] = -1
    with pytest.warns(AstropyDeprecationWarning):
        assert not is_rotation(m2)
    with pytest.warns(AstropyDeprecationWarning):
        assert is_rotation(m2, allow_improper=True)
    # and (M, 3, 3)
    n2 = np.stack((m1, m2))
    with pytest.warns(AstropyDeprecationWarning):
        assert tuple(is_rotation(n2)) == (True, False)  # (show the broadcasting)

    # Not any sort of rotation
    m3 = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    with pytest.warns(AstropyDeprecationWarning):
        assert not is_rotation(m3)
    with pytest.warns(AstropyDeprecationWarning):
        assert not is_rotation(m3, allow_improper=True)
    # and (M, 3, 3)
    n3 = np.stack((m1, m3))
    with pytest.warns(AstropyDeprecationWarning):
        assert tuple(is_rotation(n3)) == (True, False)  # (show the broadcasting)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Distribution class and associated machinery.
"""

import builtins

import numpy as np

from astropy import stats
from astropy import units as u
from astropy.utils.compat.numpycompat import NUMPY_LT_2_0

if NUMPY_LT_2_0:
    from numpy.core.multiarray import normalize_axis_index
    from numpy.lib.function_base import _parse_gufunc_signature
    from numpy.lib.stride_tricks import DummyArray
else:
    from numpy.lib._function_base_impl import _parse_gufunc_signature
    from numpy.lib._stride_tricks_impl import DummyArray
    from numpy.lib.array_utils import normalize_axis_index


from .function_helpers import FUNCTION_HELPERS

__all__ = ["Distribution"]


# we set this by hand because the symbolic expression (below) requires scipy
# SMAD_SCALE_FACTOR = 1 / scipy.stats.norm.ppf(0.75)
SMAD_SCALE_FACTOR = 1.48260221850560203193936104071326553821563720703125


class Distribution:
    """A scalar value or array values with associated uncertainty distribution.

    This object will take its exact type from whatever the ``samples``
    argument is. In general this is expected to be ``NdarrayDistribution`` for
    |ndarray| input, and, e.g., ``QuantityDistribution`` for a subclass such
    as |Quantity|. But anything compatible with `numpy.asanyarray` is possible
    (generally producing ``NdarrayDistribution``).

    See also: https://docs.astropy.org/en/stable/uncertainty/

    Parameters
    ----------
    samples : array-like
        The distribution, with sampling along the *trailing* axis. If 1D, the sole
        dimension is used as the sampling axis (i.e., it is a scalar distribution).
        If an |ndarray| or subclass, the data will not be copied unless it is not
        possible to take a view (generally, only when the strides of the last axis
        are negative).

    """

    _generated_subclasses = {}

    def __new__(cls, samples):
        if isinstance(samples, Distribution):
            samples = samples.distribution
        else:
            # We can handle strides larger than the itemsize by stride trickery,
            # but the samples do have to have positive offset from each other,
            # so if that's not the case, we make a copy.
            if not isinstance(samples, np.ndarray) or (
                samples.strides[-1] < samples.dtype.itemsize
                and samples.shape[-1:] != (1,)  # strides[-1] can be 0.
            ):
                samples = np.asanyarray(samples, order="C")
            if samples.shape == ():
                raise TypeError("Attempted to initialize a Distribution with a scalar")

        # Do the view in two stages, since for viewing as a new class, it is good
        # to have with the dtype in place (e.g., for Quantity.__array_finalize__,
        # it is needed to create the correct StructuredUnit).
        # First, get the structured dtype with a single "samples" field that holds
        # a size n_samples array of "sample" fields with individual samples.
        # We need the double indirection to be able to deal with non-contiguous samples
        # (such as result from indexing a structured array).
        # Note that for a single sample, samples.strides[-1] can be 0.
        new_dtype = cls._get_distribution_dtype(
            samples.dtype, samples.shape[-1], itemsize=samples.strides[-1]
        )
        # Create a structured array with the new dtype. We ensure we start with a
        # regular ndarray, to avoid interference between our complicated dtype and
        # something else (such as would be the case for Quantity.__array_finalize__,
        # which needs the dtype to be correct for creating a StructuredUnit).
        if samples.dtype.itemsize == samples.strides[-1]:
            # For contiguous last axis, a plain new view suffices.
            structured = samples.view(np.ndarray)
        else:
            # But if the last axis is not contiguous, we use __array_interface__ (as in
            # np.lib.stridetricks.as_strided) to create a new ndarray where the last
            # axis is covered by a void of the same size as the structured new_dtype
            # (assigning it directly does not work: possible empty fields get entries).
            interface = dict(samples.__array_interface__)
            interface["descr"] = interface["typestr"] = f"|V{new_dtype.itemsize}"
            interface["shape"] = samples.shape[:-1]
            interface["strides"] = samples.strides[:-1]
            structured = np.asarray(DummyArray(interface, base=samples))
        # Set our new structured dtype.
        structured.dtype = new_dtype
        # Get rid of trailing dimension of 1.
        structured.shape = samples.shape[:-1]

        # Now view as the Distribution subclass, and finalize based on the
        # original samples (e.g., to set the unit for QuantityDistribution).
        new_cls = cls._get_distribution_cls(type(samples))
        self = structured.view(new_cls)
        self.__array_finalize__(samples)
        return self

    @classmethod
    def _get_distribution_cls(cls, samples_cls):
        if issubclass(samples_cls, Distribution):
            return samples_cls

        new_cls = cls._generated_subclasses.get(samples_cls)
        if new_cls is None:
            # Walk through MRO and find closest base data class.
            # Note: right now, will basically always be ndarray, but
            # one could imagine needing some special care for one subclass,
            # which would then get its own entry.  E.g., if MaskedAngle
            # defined something special, then MaskedLongitude should depend
            # on it.
            for mro_item in samples_cls.__mro__:
                base_cls = cls._generated_subclasses.get(mro_item)
                if base_cls is not None:
                    break
            else:
                # Just hope that NdarrayDistribution can handle it.
                # TODO: this covers the case where a user puts in a list or so,
                # but for those one could just explicitly do something like
                # _generated_subclasses[list] = NdarrayDistribution
                return NdarrayDistribution

            # We need to get rid of the top _DistributionRepr, since we add
            # it again below (it should always be on top).
            # TODO: remove the need for _DistributionRepr by defining
            # __array_function__ and overriding array2string.
            if base_cls.__mro__[1] is not _DistributionRepr:
                # Sanity check. This should never happen!
                raise RuntimeError(
                    f"found {base_cls=}, which does not have _DistributionRepr at "
                    "__mro__[1]. Please raise an issue at "
                    "https://github.com/astropy/astropy/issues/new/choose."
                )
            # Create (and therefore register) new Distribution subclass for the
            # given samples_cls.
            new_cls = type(
                samples_cls.__name__ + "Distribution",
                (_DistributionRepr, samples_cls) + base_cls.__mro__[2:],
                {"_samples_cls": samples_cls},
            )
            cls._generated_subclasses[samples_cls] = new_cls

        return new_cls

    @classmethod
    def _get_distribution_dtype(cls, dtype, n_samples, itemsize=None):
        dtype = np.dtype(dtype)
        # If not a sample dtype already, create one with a "samples" entry that holds
        # all the samples.  Here, we create an indirection via "sample" because
        # selecting an item from a structured array will necessarily give
        # non-consecutive samples, and these can only be dealt with samples that are
        # larger than a single element (plus stride tricks; see __new__). For those
        # cases, itemsize larger than dtype.itemsize will be passed in (note that for
        # the n_sample=1 case, itemsize can be 0; like itemsize=None, this will be dealt
        # with by "or dtype.itemsize" below).
        if dtype.names != ("samples",):
            sample_dtype = np.dtype(
                dict(
                    names=["sample"],
                    formats=[dtype],
                    itemsize=itemsize or dtype.itemsize,
                )
            )
            dtype = np.dtype([("samples", sample_dtype, (n_samples,))])
        return dtype

    @property
    def distribution(self):
        return self["samples"]["sample"]

    @property
    def dtype(self):
        return super().dtype["samples"].base["sample"]

    @dtype.setter
    def dtype(self, dtype):
        dtype = self._get_distribution_dtype(
            dtype, self.n_samples, itemsize=super().dtype["samples"].base.itemsize
        )
        super(Distribution, self.__class__).dtype.__set__(self, dtype)

    def astype(self, dtype, *args, **kwargs):
        dtype = self._get_distribution_dtype(dtype, self.n_samples)
        return super().astype(dtype, *args, **kwargs)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        converted = []
        outputs = kwargs.pop("out", None)
        if outputs:
            kwargs["out"] = tuple(
                (output.distribution if isinstance(output, Distribution) else output)
                for output in outputs
            )
            if ufunc.nout == 1:
                outputs = outputs[0]

        axis = kwargs.get("axis")
        keepdims = kwargs.get("keepdims", False)
        if method in {"reduce", "accumulate", "reduceat"}:
            if axis is None:
                assert isinstance(inputs[0], Distribution)
                kwargs["axis"] = tuple(range(inputs[0].ndim))

        for input_ in inputs:
            # For any input that is not a Distribution, we add an axis at the
            # end, to allow proper broadcasting with the distributions.
            if isinstance(input_, Distribution):
                converted.append(input_.distribution)
            else:
                shape = getattr(input_, "shape", ())
                if shape:
                    converted.append(input_[..., np.newaxis])
                else:
                    converted.append(input_)

        if ufunc.signature:
            # We're dealing with a gufunc, which may add an extra axis.
            # Ignore axes keyword for now...  TODO: remove this limitation.
            if "axes" in kwargs:
                raise NotImplementedError(
                    "Distribution does not yet support gufunc calls with 'axes'."
                )
            # Parse signature with private numpy function. Note it
            # cannot handle spaces in tuples, so remove those.
            in_sig, out_sig = _parse_gufunc_signature(ufunc.signature.replace(" ", ""))
            ncore_in = [len(sig) for sig in in_sig]
            ncore_out = [len(sig) + (1 if keepdims else 0) for sig in out_sig]
            if ufunc.nout == 1:
                ncore_out = ncore_out[0]
            if axis is None:
                axis = -1
            converted = [
                np.moveaxis(conv, -1, normalize_axis_index(axis, conv.ndim) - ncore)
                if ncore and getattr(conv, "shape", ())
                else conv
                for conv, ncore in zip(converted, ncore_in)
            ]
        else:
            ncore_out = None

        results = getattr(ufunc, method)(*converted, **kwargs)

        return self._result_as_distribution(results, outputs, ncore_out, axis)

    def __array_function__(self, function, types, args, kwargs):
        # TODO: go through functions systematically to see which ones
        # work and/or can be supported.
        if function in FUNCTION_HELPERS:
            function_helper = FUNCTION_HELPERS[function]
            try:
                args, kwargs, out = function_helper(*args, **kwargs)
            except NotImplementedError:
                return self._not_implemented_or_raise(function, types)

            result = super().__array_function__(function, types, args, kwargs)
            # Fall through to return section, wrapping distributions, unless
            # indicated otherwise.  TODO: use a less hacky solution?
            if out is True:
                return result

        else:  # pragma: no cover
            # By default, just pass it through for now.
            return super().__array_function__(function, types, args, kwargs)

        # We're done if the result was NotImplemented, which can happen
        # if other inputs/outputs override __array_function__;
        # hopefully, they can then deal with us.
        if result is NotImplemented:
            return NotImplemented

        return self._result_as_distribution(result, out=out)

    def _result_as_distribution(self, result, out, ncore_out=None, axis=None):
        """Turn result into a distribution.

        If no output is given, it will create a Distribution from the array,
        If an output is given, it should be fine as is.

        Parameters
        ----------
        result : ndarray or tuple thereof
            Array(s) which need to be turned into Distribution.
        out : Distribution, tuple of Distribution or None
            Possible output |Distribution|. Should be `None` or a tuple if result
            is a tuple.
        ncore_out: int or tuple thereof
            The number of core dimensions for the output array for a gufunc.  This
            is used to determine which axis should be used for the samples.
        axis: int or None
            The axis a gufunc operated on.  Used only if ``ncore_out`` is given.

        Returns
        -------
        out : Distribution
        """
        if isinstance(result, (tuple, list)):
            if out is None:
                out = (None,) * len(result)
            if ncore_out is None:
                ncore_out = (None,) * len(result)
            # Some np.linalg functions return namedtuple, which is handy to access
            # elements by name, but cannot be directly initialized with an iterator.
            result_cls = getattr(result, "_make", result.__class__)
            return result_cls(
                self._result_as_distribution(result_, out_, axis=axis, ncore_out=ncore)
                for (result_, out_, ncore) in zip(result, out, ncore_out)
            )

        if out is None:
            # Turn the result into a Distribution if needed.
            if not isinstance(result, Distribution) and getattr(result, "shape", ()):
                if ncore_out is not None:
                    result = np.moveaxis(
                        result, normalize_axis_index(axis, result.ndim) - ncore_out, -1
                    )
                return Distribution(result)
            else:
                return result
        else:
            # TODO: remove this sanity check once test cases are more complete.
            assert isinstance(out, Distribution)
            return out

    def _not_implemented_or_raise(self, function, types):
        # Our function helper or dispatcher found that the function does not
        # work with Distribution.  In principle, there may be another class that
        # knows what to do with us, for which we should return NotImplemented.
        # But if there is ndarray (or a non-Distribution subclass of it) around,
        # it quite likely coerces, so we should just break.
        if any(
            issubclass(t, np.ndarray) and not issubclass(t, Distribution) for t in types
        ):
            raise TypeError(
                f"the Distribution implementation cannot handle {function} "
                "with the given arguments."
            ) from None
        else:
            return NotImplemented

    # Override __eq__ and __ne__ to pass on directly to the ufunc since
    # otherwise comparisons with non-distributions do not work (but
    # deferring if other defines __array_ufunc__ = None -- see
    # numpy/core/src/common/binop_override.h for the logic; we assume we
    # will never deal with __array_priority__ any more).  Note: there is no
    # problem for other comparisons, since for those, structured arrays are
    # not treated differently in numpy/core/src/multiarray/arrayobject.c.
    def __eq__(self, other):
        if getattr(other, "__array_ufunc__", False) is None:
            return NotImplemented
        return np.equal(self, other)

    def __ne__(self, other):
        if getattr(other, "__array_ufunc__", False) is None:
            return NotImplemented
        return np.not_equal(self, other)

    @property
    def n_samples(self):
        """
        The number of samples of this distribution.  A single `int`.
        """
        return super().dtype["samples"].shape[0]

    def pdf_mean(self, dtype=None, out=None):
        """
        The mean of this distribution.

        Arguments are as for `numpy.mean`.
        """
        return self.distribution.mean(axis=-1, dtype=dtype, out=out)

    def pdf_std(self, dtype=None, out=None, ddof=0):
        """
        The standard deviation of this distribution.

        Arguments are as for `numpy.std`.
        """
        return self.distribution.std(axis=-1, dtype=dtype, out=out, ddof=ddof)

    def pdf_var(self, dtype=None, out=None, ddof=0):
        """
        The variance of this distribution.

        Arguments are as for `numpy.var`.
        """
        return self.distribution.var(axis=-1, dtype=dtype, out=out, ddof=ddof)

    def pdf_median(self, out=None):
        """
        The median of this distribution.

        Parameters
        ----------
        out : array, optional
            Alternative output array in which to place the result. It must
            have the same shape and buffer length as the expected output,
            but the type (of the output) will be cast if necessary.
        """
        return np.median(self.distribution, axis=-1, out=out)

    def pdf_mad(self, out=None):
        """
        The median absolute deviation of this distribution.

        Parameters
        ----------
        out : array, optional
            Alternative output array in which to place the result. It must
            have the same shape and buffer length as the expected output,
            but the type (of the output) will be cast if necessary.
        """
        median = self.pdf_median(out=out)
        absdiff = np.abs(self - median)
        return np.median(
            absdiff.distribution, axis=-1, out=median, overwrite_input=True
        )

    def pdf_smad(self, out=None):
        """
        The median absolute deviation of this distribution rescaled to match the
        standard deviation for a normal distribution.

        Parameters
        ----------
        out : array, optional
            Alternative output array in which to place the result. It must
            have the same shape and buffer length as the expected output,
            but the type (of the output) will be cast if necessary.
        """
        result = self.pdf_mad(out=out)
        result *= SMAD_SCALE_FACTOR
        return result

    def pdf_percentiles(self, percentile, **kwargs):
        """
        Compute percentiles of this Distribution.

        Parameters
        ----------
        percentile : float or array of float or `~astropy.units.Quantity`
            The desired percentiles of the distribution (i.e., on [0,100]).
            `~astropy.units.Quantity` will be converted to percent, meaning
            that a ``dimensionless_unscaled`` `~astropy.units.Quantity` will
            be interpreted as a quantile.

        Additional keywords are passed into `numpy.percentile`.

        Returns
        -------
        percentiles : `~astropy.units.Quantity` ['dimensionless']
            The ``fracs`` percentiles of this distribution.
        """
        percentile = u.Quantity(percentile, u.percent).value
        percs = np.percentile(self.distribution, percentile, axis=-1, **kwargs)
        # numpy.percentile strips units for unclear reasons, so we have to make
        # a new object with units
        if hasattr(self.distribution, "_new_view"):
            return self.distribution._new_view(percs)
        else:
            return percs

    def pdf_histogram(self, **kwargs):
        """
        Compute histogram over the samples in the distribution.

        Parameters
        ----------
        All keyword arguments are passed into `astropy.stats.histogram`. Note
        That some of these options may not be valid for some multidimensional
        distributions.

        Returns
        -------
        hist : array
            The values of the histogram. Trailing dimension is the histogram
            dimension.
        bin_edges : array of dtype float
            Return the bin edges ``(length(hist)+1)``. Trailing dimension is the
            bin histogram dimension.
        """
        distr = self.distribution
        raveled_distr = distr.reshape(distr.size // distr.shape[-1], distr.shape[-1])

        nhists = []
        bin_edges = []
        for d in raveled_distr:
            nhist, bin_edge = stats.histogram(d, **kwargs)
            nhists.append(nhist)
            bin_edges.append(bin_edge)

        nhists = np.array(nhists)
        nh_shape = self.shape + (nhists.size // self.size,)
        bin_edges = np.array(bin_edges)
        be_shape = self.shape + (bin_edges.size // self.size,)
        return nhists.reshape(nh_shape), bin_edges.reshape(be_shape)


class ScalarDistribution(Distribution, np.void):
    """Scalar distribution.

    This class mostly exists to make `~numpy.array2print` possible for
    all subclasses.  It is a scalar element, still with n_samples samples.
    """


class ArrayDistribution(Distribution, np.ndarray):
    # This includes the important override of view and __getitem__
    # which are needed for all ndarray subclass Distributions, but not
    # for the scalar one.
    _samples_cls = np.ndarray

    # Override view so that we stay a Distribution version of the new type.
    def view(self, dtype=None, type=None):
        """New view of array with the same data.

        Like `~numpy.ndarray.view` except that the result will always be a new
        `~astropy.uncertainty.Distribution` instance.  If the requested
        ``type`` is a `~astropy.uncertainty.Distribution`, then no change in
        ``dtype`` is allowed.

        """
        if type is None:
            if isinstance(dtype, builtins.type) and issubclass(dtype, np.ndarray):
                type = self._get_distribution_cls(dtype)
                dtype = None
            else:
                type = self.__class__
        else:
            type = self._get_distribution_cls(type)

        type = self._get_distribution_cls(type)
        if dtype is None:
            return super().view(type=type)

        dtype = np.dtype(dtype)
        if dtype == self.dtype:
            return super().view(type=type)

        if dtype.names == ("samples",):
            # Assume the user knows what they are doing.
            return super().view(dtype, type)

        if dtype.shape == () and dtype.itemsize == self.dtype.itemsize:
            dtype = self._get_distribution_dtype(
                dtype,
                self.n_samples,
                itemsize=super(Distribution, self).dtype["samples"].base.itemsize,
            )
            return super().view(dtype, type)

        samples_cls = type._samples_cls
        if dtype.itemsize == self.dtype.itemsize:
            distr = self.distribution
            distr_view = distr.view(dtype, samples_cls)
            # This does not necessarily leave the sample axis in the right place.
            return Distribution(np.moveaxis(distr_view, distr.ndim - 1, -1))
        elif dtype.itemsize == super(Distribution, self).dtype["samples"].base.itemsize:
            distr = np.moveaxis(self.distribution, -1, -2)
            distr_view = distr.view(dtype, samples_cls).squeeze(-1)
            return Distribution(distr_view)
        else:
            raise ValueError(
                f"{self.__class__} can only be viewed with a dtype with "
                "itemsize {self.strides[-1]} or {self.dtype.itemsize}"
            )

    @property
    def distribution(self):
        # Like in the creation, we go through an ndarray to ensure we have our
        # actual dtype and to avoid entering, e.g., Quantity.__getitem__, which
        # would give problems with units.
        structured = super().view(np.ndarray)
        distribution = structured["samples"]["sample"].view(self._samples_cls)
        distribution.__array_finalize__(self)
        return distribution

    def __getitem__(self, item):
        if isinstance(item, str):
            # "samples" should always get back to the samples class.
            if item == "samples":
                return self.distribution
            else:
                # Hard to get this right directly, so instead get item from the
                # distribution, and create a new instance.  We move the sample axis to
                # the end to ensure the order is right for possible subarrays.
                return Distribution(np.moveaxis(self.distribution[item], self.ndim, -1))

        if isinstance(item, Distribution):
            # Required for in-place operations like dist[dist < 0] += 360.
            return self.distribution[item.distribution]

        result = super().__getitem__(item)
        if isinstance(result, np.void):
            return result.view((ScalarDistribution, result.dtype))
        else:
            return result

    def __setitem__(self, item, value):
        if isinstance(item, Distribution):
            # Support operations like dist[dist < 0] = 0.
            self.distribution[item.distribution] = value
            return

        if isinstance(item, str):
            if item == "samples":
                self.distribution[()] = value
                return

            # Get a view of this item (non-trivial; see above).
            self = self[item]
            item = ()

        if not isinstance(value, Distribution):
            # If value is not already a Distribution, first make it an array
            # to help interpret possible structured dtype, and then turn it
            # into a Distribution with n_samples=1 (which will broadcast).
            value = np.asanyarray(value, dtype=self.dtype)
            value = Distribution(value[..., np.newaxis])

        super().__setitem__(item, value)


class _DistributionRepr:
    def __repr__(self):
        reprarr = repr(self.distribution)
        if reprarr.endswith(">"):
            firstspace = reprarr.find(" ")
            reprarr = reprarr[firstspace + 1 : -1]  # :-1] removes the ending '>'
            return (
                f"<{self.__class__.__name__} {reprarr} with n_samples={self.n_samples}>"
            )
        else:  # numpy array-like
            firstparen = reprarr.find("(")
            reprarr = reprarr[firstparen:]
            return f"{self.__class__.__name__}{reprarr} with n_samples={self.n_samples}"

    def __str__(self):
        distrstr = str(self.distribution)
        toadd = f" with n_samples={self.n_samples}"
        return distrstr + toadd

    def _repr_latex_(self):
        if hasattr(self.distribution, "_repr_latex_"):
            superlatex = self.distribution._repr_latex_()
            toadd = rf", \; n_{{\rm samp}}={self.n_samples}"
            return superlatex[:-1] + toadd + superlatex[-1]
        else:
            return None


class NdarrayDistribution(_DistributionRepr, ArrayDistribution):
    pass


# Ensure our base NdarrayDistribution is known.
Distribution._generated_subclasses[np.ndarray] = NdarrayDistribution



from abc import abstractmethod
from math import exp, log

import numpy as np
from numpy.typing import ArrayLike

from astropy.cosmology._src.scipy_compat import quad
from astropy.cosmology._src.typing import FArray
from astropy.cosmology._src.utils import aszarr, deprecated_keywords
from astropy.units import Quantity


class DarkEnergyComponent:
    # Subclasses should use `Parameter` to make this a parameter of the cosmology.
    Ode0: float
    """Omega dark energy; dark energy density/critical density at z=0."""

    @abstractmethod
    @deprecated_keywords("z", since="7.0")
    def w(self, z: Quantity | ArrayLike) -> FArray:
        r"""The dark energy equation of state.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        w : ndarray or float
            The dark energy equation of state.
            `float` if scalar input.

        Notes
        -----
        The dark energy equation of state is defined as
        :math:`w(z) = P(z)/\rho(z)`, where :math:`P(z)` is the pressure at
        redshift z and :math:`\rho(z)` is the density at redshift z, both in
        units where c=1.

        This must be overridden by subclasses.
        """
        raise NotImplementedError("w(z) is not implemented")

    def _w_integrand(self, ln1pz: float | FArray, /) -> FArray:
        """Internal convenience function for w(z) integral (eq. 5 of [1]_).

        Parameters
        ----------
        ln1pz : `~numbers.Number` or scalar ndarray, positional-only
            Assumes scalar input, since this should only be called inside an
            integral.

            .. versionchanged:: 7.0
                The argument is positional-only.

        References
        ----------
        .. [1] Linder, E. (2003). Exploring the Expansion History of the
               Universe. Phys. Rev. Lett., 90, 091301.
        """
        return 1.0 + self.w(exp(ln1pz) - 1.0)

    @deprecated_keywords("z", since="7.0")
    def de_density_scale(self, z: Quantity | ArrayLike) -> FArray:
        r"""Evaluates the redshift dependence of the dark energy density.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        I : ndarray or float
            The scaling of the energy density of dark energy with redshift.
            Returns `float` if the input is scalar.

        Notes
        -----
        The scaling factor, I, is defined by :math:`\rho(z) = \rho_0 I`,
        and is given by

        .. math::

           I = \exp \left( 3 \int_{a}^1 \frac{ da^{\prime} }{ a^{\prime} }
                          \left[ 1 + w\left( a^{\prime} \right) \right] \right)

        The actual integral used is rewritten from [1]_ to be in terms of z.

        It will generally helpful for subclasses to overload this method if
        the integral can be done analytically for the particular dark
        energy equation of state that they implement.

        References
        ----------
        .. [1] Linder, E. (2003). Exploring the Expansion History of the
               Universe. Phys. Rev. Lett., 90, 091301.
        """
        # This allows for an arbitrary w(z) following eq (5) of
        # Linder 2003, PRL 90, 91301.  The code here evaluates
        # the integral numerically.  However, most popular
        # forms of w(z) are designed to make this integral analytic,
        # so it is probably a good idea for subclasses to overload this
        # method if an analytic form is available.
        z = aszarr(z)
        ival = (
            quad(self._w_integrand, 0, log(z + 1.0))[0]  # scalar
            if z.ndim == 0
            else np.asarray([quad(self._w_integrand, 0, log(1 + _z))[0] for _z in z])
        )
        return np.exp(3 * ival)

    @deprecated_keywords("z", since="7.0")
    def Ode(self, z: Quantity | ArrayLike) -> FArray:
        """Return the density parameter for dark energy at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        Ode : ndarray
            The density of dark energy relative to the critical density at each
            redshift.
        """
        z = aszarr(z)
        if self.Ode0 == 0:  # Common enough to be worth checking explicitly
            return np.zeros_like(z)
        # Ensure self.inv_efunc is implemented by the main class
        if not hasattr(self, "inv_efunc") or not callable(self.inv_efunc):
            msg = "The main class must implement an 'inv_efunc(z)' method."
            raise NotImplementedError(msg)
        return self.Ode0 * self.de_density_scale(z) * self.inv_efunc(z) ** 2



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Utilities for console input and output.
"""

import codecs
import locale
import math
import multiprocessing
import os
import struct
import sys
import threading
import time
from shutil import get_terminal_size

# concurrent.futures imports moved inside functions using them to avoid
# import failure when running in pyodide/Emscripten

try:
    import fcntl
    import signal
    import termios

    _CAN_RESIZE_TERMINAL = True
except ImportError:
    _CAN_RESIZE_TERMINAL = False

import numpy as np

from astropy import conf
from astropy.utils.compat.optional_deps import (
    HAS_IPYKERNEL,
    HAS_IPYTHON,
    HAS_IPYWIDGETS,
)

from .decorators import classproperty, deprecated

__all__ = [
    "ProgressBar",
    "ProgressBarOrSpinner",
    "Spinner",
    "color_print",
    "human_file_size",
    "human_time",
    "isatty",
    "print_code_line",
    "terminal_size",
]


class _IPython:
    """Singleton class given access to IPython streams, etc."""

    @classproperty
    def OutStream(cls):
        if not hasattr(cls, "_OutStream"):
            if HAS_IPYKERNEL:
                from ipykernel.iostream import OutStream

                cls._OutStream = OutStream
            else:
                cls._OutStream = None

        return cls._OutStream

    @classproperty
    def ipyio(cls):
        if not hasattr(cls, "_ipyio"):
            if HAS_IPYTHON:
                from IPython.utils import io

                cls._ipyio = io
            else:
                cls._ipyio = None
        return cls._ipyio


def isatty(file):
    """
    Returns `True` if ``file`` is a tty.

    Most built-in Python file-like objects have an `isatty` member,
    but some user-defined types may not, so this assumes those are not
    ttys.
    """
    if (
        multiprocessing.current_process().name != "MainProcess"
        or threading.current_thread().name != "MainThread"
    ):
        return False

    if hasattr(file, "isatty"):
        return file.isatty()

    if _IPython.OutStream is None or (not isinstance(file, _IPython.OutStream)):
        return False

    return getattr(file, "name", None) == "stdout"


@deprecated("6.1", alternative="shutil.get_terminal_size")
def terminal_size(file=None):
    """
    Returns a tuple (height, width) containing the height and width of
    the terminal.

    This function will look for the width in height in multiple areas
    before falling back on the width and height in astropy's
    configuration.
    """
    if file is None:
        file = sys.stdout

    try:
        s = struct.pack("HHHH", 0, 0, 0, 0)
        x = fcntl.ioctl(file, termios.TIOCGWINSZ, s)
        (lines, width, xpixels, ypixels) = struct.unpack("HHHH", x)
        if lines > 12:
            lines -= 6
        if width > 10:
            width -= 1
        if lines <= 0 or width <= 0:
            raise Exception("unable to get terminal size")
        return (lines, width)
    except Exception:
        try:
            # see if POSIX standard variables will work
            return (int(os.environ.get("LINES")), int(os.environ.get("COLUMNS")))
        except TypeError:
            # fall back on configuration variables, or if not
            # set, (25, 80)
            lines = conf.max_lines
            width = conf.max_width
            if lines is None:
                lines = 25
            if width is None:
                width = 80
            return lines, width


def _color_text(text, color):
    """Returns a string wrapped in ANSI color codes for coloring the text in a terminal.

    ::

        colored_text = color_text('Here is a message', 'blue')

    This won't actually effect the text until it is printed to the
    terminal.

    Parameters
    ----------
    text : str
        The string to return, bounded by the color codes.
    color : str
        An ANSI terminal color name. Must be one of:
        black, red, green, brown, blue, magenta, cyan, lightgrey,
        default, darkgrey, lightred, lightgreen, yellow, lightblue,
        lightmagenta, lightcyan, white, or '' (the empty string).
    """
    color_mapping = {
        "black": "0;30",
        "red": "0;31",
        "green": "0;32",
        "brown": "0;33",
        "blue": "0;34",
        "magenta": "0;35",
        "cyan": "0;36",
        "lightgrey": "0;37",
        "default": "0;39",
        "darkgrey": "1;30",
        "lightred": "1;31",
        "lightgreen": "1;32",
        "yellow": "1;33",
        "lightblue": "1;34",
        "lightmagenta": "1;35",
        "lightcyan": "1;36",
        "white": "1;37",
    }

    if sys.platform == "win32" and _IPython.OutStream is None:
        # On Windows do not colorize text unless in IPython
        return text

    color_code = color_mapping.get(color, "0;39")
    return f"\033[{color_code}m{text}\033[0m"


def _write_with_fallback(s, write, fileobj):
    """Write the supplied string with the given write function like
    ``write(s)``, but use a writer for the locale's preferred encoding in case
    of a UnicodeEncodeError.  Failing that attempt to write with 'utf-8' or
    'latin-1'.
    """
    try:
        write(s)
        return write
    except UnicodeEncodeError:
        # Let's try the next approach...
        pass

    enc = locale.getpreferredencoding()
    try:
        Writer = codecs.getwriter(enc)
    except LookupError:
        Writer = codecs.getwriter("utf-8")

    f = Writer(fileobj)
    write = f.write

    try:
        write(s)
        return write
    except UnicodeEncodeError:
        Writer = codecs.getwriter("latin-1")
        f = Writer(fileobj)
        write = f.write

    # If this doesn't work let the exception bubble up; I'm out of ideas
    write(s)
    return write


def color_print(*args, end="\n", **kwargs):
    """
    Prints colors and styles to the terminal uses ANSI escape
    sequences.

    ::

       color_print('This is the color ', 'default', 'GREEN', 'green')

    Parameters
    ----------
    positional args : str
        The positional arguments come in pairs (*msg*, *color*), where
        *msg* is the string to display and *color* is the color to
        display it in.

        *color* is an ANSI terminal color name.  Must be one of:
        black, red, green, brown, blue, magenta, cyan, lightgrey,
        default, darkgrey, lightred, lightgreen, yellow, lightblue,
        lightmagenta, lightcyan, white, or '' (the empty string).

    file : :term:`file-like (writeable)`, optional
        Where to write to.  Defaults to `sys.stdout`.  If file is not
        a tty (as determined by calling its `isatty` member, if one
        exists), no coloring will be included.

    end : str, optional
        The ending of the message.  Defaults to ``\\n``.  The end will
        be printed after resetting any color or font state.
    """
    file = kwargs.get("file", sys.stdout)

    write = file.write
    if isatty(file) and conf.use_color:
        for i in range(0, len(args), 2):
            msg = args[i]
            if i + 1 == len(args):
                color = ""
            else:
                color = args[i + 1]

            if color:
                msg = _color_text(msg, color)

            # Some file objects support writing unicode sensibly on some Python
            # versions; if this fails try creating a writer using the locale's
            # preferred encoding. If that fails too give up.

            write = _write_with_fallback(msg, write, file)

        write(end)
    else:
        for i in range(0, len(args), 2):
            msg = args[i]
            write(msg)
        write(end)


def human_time(seconds):
    """
    Returns a human-friendly time string that is always exactly 6
    characters long.

    Depending on the number of seconds given, can be one of::

        1w 3d
        2d 4h
        1h 5m
        1m 4s
          15s

    Will be in color if console coloring is turned on.

    Parameters
    ----------
    seconds : int
        The number of seconds to represent

    Returns
    -------
    time : str
        A human-friendly representation of the given number of seconds
        that is always exactly 6 characters.
    """
    units = [
        ("y", 60 * 60 * 24 * 7 * 52),
        ("w", 60 * 60 * 24 * 7),
        ("d", 60 * 60 * 24),
        ("h", 60 * 60),
        ("m", 60),
        ("s", 1),
    ]

    seconds = int(seconds)

    if seconds < 60:
        return f"   {seconds:2d}s"
    for i in range(len(units) - 1):
        unit1, limit1 = units[i]
        unit2, limit2 = units[i + 1]
        if seconds >= limit1:
            return (
                f"{seconds // limit1:2d}{unit1}{(seconds % limit1) // limit2:2d}{unit2}"
            )
    return "  ~inf"


def human_file_size(size):
    """
    Returns a human-friendly string representing a file size
    that is 2-4 characters long.

    For example, depending on the number of bytes given, can be one
    of::

        256b
        64k
        1.1G

    Parameters
    ----------
    size : int
        The size of the file (in bytes)

    Returns
    -------
    size : str
        A human-friendly representation of the size of the file
    """
    if hasattr(size, "unit"):
        # Import units only if necessary because the import takes a
        # significant time [#4649]
        from astropy import units as u

        size = u.Quantity(size, u.byte).value

    suffixes = " kMGTPEZY"
    if size == 0:
        num_scale = 0
    else:
        num_scale = math.floor(math.log(size) / math.log(1000))
    if num_scale > 7:
        suffix = "?"
    else:
        suffix = suffixes[num_scale]
    num_scale = int(math.pow(1000, num_scale))
    value = size / num_scale
    str_value = str(value)
    if suffix == " ":
        str_value = str_value[: str_value.index(".")]
    elif str_value[2] == ".":
        str_value = str_value[:2]
    else:
        str_value = str_value[:3]
    return f"{str_value:>3s}{suffix}"


class _mapfunc:
    """
    A function wrapper to support ProgressBar.map().
    """

    def __init__(self, func):
        self._func = func

    def __call__(self, i_arg):
        i, arg = i_arg
        return i, self._func(arg)


class ProgressBar:
    """
    A class to display a progress bar in the terminal.

    It is designed to be used either with the ``with`` statement::

        with ProgressBar(len(items)) as bar:
            for item in enumerate(items):
                bar.update()

    or as a generator::

        for item in ProgressBar(items):
            item.process()
    """

    def __init__(self, total_or_items, ipython_widget=False, file=None):
        """
        Parameters
        ----------
        total_or_items : int or sequence
            If an int, the number of increments in the process being
            tracked.  If a sequence, the items to iterate over.

        ipython_widget : bool, optional
            If `True`, the progress bar will display as an IPython
            notebook widget.

        file : :term:`file-like (writeable)`, optional
            The file to write the progress bar to.  Defaults to
            `sys.stdout`.  If ``file`` is not a tty (as determined by
            calling its `isatty` member, if any, or special case hacks
            to detect the IPython console), the progress bar will be
            completely silent.
        """
        if file is None:
            file = sys.stdout

        if not ipython_widget and not isatty(file):
            self.update = self._silent_update
            self._silent = True
        else:
            self._silent = False

        if np.iterable(total_or_items):
            self._items = iter(total_or_items)
            self._total = len(total_or_items)
        else:
            try:
                self._total = int(total_or_items)
            except TypeError:
                raise TypeError("First argument must be int or sequence")
            else:
                self._items = iter(range(self._total))

        self._file = file
        self._start_time = time.time()
        self._human_total = human_file_size(self._total)
        self._ipython_widget = ipython_widget

        self._signal_set = False
        if not ipython_widget:
            self._should_handle_resize = _CAN_RESIZE_TERMINAL and self._file.isatty()
            self._handle_resize()
            if self._should_handle_resize:
                signal.signal(signal.SIGWINCH, self._handle_resize)
                self._signal_set = True

        self.update(0)

    def _handle_resize(self, signum=None, frame=None):
        terminal_width = get_terminal_size().columns
        self._bar_length = terminal_width - 37

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if not self._silent:
            if exc_type is None:
                self.update(self._total)
            self._file.write("\n")
            self._file.flush()
            if self._signal_set:
                signal.signal(signal.SIGWINCH, signal.SIG_DFL)

    def __iter__(self):
        return self

    def __next__(self):
        try:
            rv = next(self._items)
        except StopIteration:
            self.__exit__(None, None, None)
            raise
        else:
            self.update()
            return rv

    def update(self, value=None):
        """
        Update progress bar via the console or notebook accordingly.
        """
        # Update self.value
        if value is None:
            value = self._current_value + 1
        self._current_value = value

        # Choose the appropriate environment
        if self._ipython_widget:
            self._update_ipython_widget(value)
        else:
            self._update_console(value)

    def _update_console(self, value=None):
        """
        Update the progress bar to the given value (out of the total
        given to the constructor).
        """
        if self._total == 0:
            frac = 1.0
        else:
            frac = float(value) / float(self._total)

        file = self._file
        write = file.write

        if frac > 1:
            bar_fill = int(self._bar_length)
        else:
            bar_fill = int(float(self._bar_length) * frac)
        write("\r|")
        color_print("=" * bar_fill, "blue", file=file, end="")
        if bar_fill < self._bar_length:
            color_print(">", "green", file=file, end="")
            write("-" * (self._bar_length - bar_fill - 1))
        write("|")

        if value >= self._total:
            t = time.time() - self._start_time
            prefix = "     "
        elif value <= 0:
            t = None
            prefix = ""
        else:
            t = ((time.time() - self._start_time) * (1.0 - frac)) / frac
            prefix = " ETA "
        write(f" {human_file_size(value):>4s}/{self._human_total:>4s}")
        write(f" ({frac:>6.2%})")
        write(prefix)
        if t is not None:
            write(human_time(t))
        self._file.flush()

    def _update_ipython_widget(self, value=None):
        """
        Update the progress bar to the given value (out of a total
        given to the constructor).

        This method is for use in the IPython notebook 2+.
        """
        # Create and display an empty progress bar widget,
        # if none exists.

        if not hasattr(self, "_widget"):
            # Import only if an IPython widget, i.e., widget in iPython NB
            if not HAS_IPYWIDGETS:
                raise ModuleNotFoundError("ipywidgets is not installed")

            from ipywidgets import widgets

            self._widget = widgets.FloatProgress()
            from IPython.display import display

            display(self._widget)
            self._widget.value = 0

        # Calculate percent completion, and update progress bar
        frac = value / self._total
        self._widget.value = frac * 100
        self._widget.description = f" ({frac:>6.2%})"

    def _silent_update(self, value=None):
        pass

    @classmethod
    def map(
        cls,
        function,
        items,
        multiprocess=False,
        file=None,
        step=100,
        ipython_widget=False,
        multiprocessing_start_method=None,
    ):
        """Map function over items while displaying a progress bar with percentage complete.

        The map operation may run in arbitrary order on the items, but the results are
        returned in sequential order.

        ::

            def work(i):
                print(i)

            ProgressBar.map(work, range(50))

        Parameters
        ----------
        function : function
            Function to call for each step

        items : sequence
            Sequence where each element is a tuple of arguments to pass to
            *function*.

        multiprocess : bool, int, optional
            If `True`, use the `multiprocessing` module to distribute each task
            to a different processor core. If a number greater than 1, then use
            that number of cores.

        ipython_widget : bool, optional
            If `True`, the progress bar will display as an IPython
            notebook widget.

        file : :term:`file-like (writeable)`, optional
            The file to write the progress bar to.  Defaults to
            `sys.stdout`.  If ``file`` is not a tty (as determined by
            calling its `isatty` member, if any), the scrollbar will
            be completely silent.

        step : int, optional
            Update the progress bar at least every *step* steps (default: 100).
            If ``multiprocess`` is `True`, this will affect the size
            of the chunks of ``items`` that are submitted as separate tasks
            to the process pool.  A large step size may make the job
            complete faster if ``items`` is very long.

        multiprocessing_start_method : str, optional
            Useful primarily for testing; if in doubt leave it as the default.
            When using multiprocessing, certain anomalies occur when starting
            processes with the "spawn" method (the only option on Windows);
            other anomalies occur with the "fork" method (the default on
            Linux).
        """
        if multiprocess:
            function = _mapfunc(function)
            items = list(enumerate(items))

        results = cls.map_unordered(
            function,
            items,
            multiprocess=multiprocess,
            file=file,
            step=step,
            ipython_widget=ipython_widget,
            multiprocessing_start_method=multiprocessing_start_method,
        )

        if multiprocess:
            _, results = zip(*sorted(results))
            results = list(results)

        return results

    @classmethod
    def map_unordered(
        cls,
        function,
        items,
        multiprocess=False,
        file=None,
        step=100,
        ipython_widget=False,
        multiprocessing_start_method=None,
    ):
        """Map function over items, reporting the progress.

        Does a `map` operation while displaying a progress bar with
        percentage complete. The map operation may run on arbitrary order
        on the items, and the results may be returned in arbitrary order.

        ::

            def work(i):
                print(i)

            ProgressBar.map(work, range(50))

        Parameters
        ----------
        function : function
            Function to call for each step

        items : sequence
            Sequence where each element is a tuple of arguments to pass to
            *function*.

        multiprocess : bool, int, optional
            If `True`, use the `multiprocessing` module to distribute each task
            to a different processor core. If a number greater than 1, then use
            that number of cores.

        ipython_widget : bool, optional
            If `True`, the progress bar will display as an IPython
            notebook widget.

        file : :term:`file-like (writeable)`, optional
            The file to write the progress bar to.  Defaults to
            `sys.stdout`.  If ``file`` is not a tty (as determined by
            calling its `isatty` member, if any), the scrollbar will
            be completely silent.

        step : int, optional
            Update the progress bar at least every *step* steps (default: 100).
            If ``multiprocess`` is `True`, this will affect the size
            of the chunks of ``items`` that are submitted as separate tasks
            to the process pool.  A large step size may make the job
            complete faster if ``items`` is very long.

        multiprocessing_start_method : str, optional
            Useful primarily for testing; if in doubt leave it as the default.
            When using multiprocessing, certain anomalies occur when starting
            processes with the "spawn" method (the only option on Windows);
            other anomalies occur with the "fork" method (the default on
            Linux).
        """
        # concurrent.futures import here to avoid import failure when running
        # in pyodide/Emscripten
        from concurrent.futures import ProcessPoolExecutor, as_completed

        results = []

        if file is None:
            file = sys.stdout

        with cls(len(items), ipython_widget=ipython_widget, file=file) as bar:
            if bar._ipython_widget:
                chunksize = step
            else:
                default_step = max(int(float(len(items)) / bar._bar_length), 1)
                chunksize = min(default_step, step)
            if not multiprocess or multiprocess < 1:
                for i, item in enumerate(items):
                    results.append(function(item))
                    if (i % chunksize) == 0:
                        bar.update(i)
            else:
                ctx = multiprocessing.get_context(multiprocessing_start_method)
                kwargs = dict(mp_context=ctx)

                with ProcessPoolExecutor(
                    max_workers=(
                        int(multiprocess) if multiprocess is not True else None
                    ),
                    **kwargs,
                ) as p:
                    for i, f in enumerate(
                        as_completed(p.submit(function, item) for item in items)
                    ):
                        bar.update(i)
                        results.append(f.result())

        return results


class Spinner:
    """
    A class to display a spinner in the terminal.

    It is designed to be used with the ``with`` statement::

        with Spinner("Reticulating splines", "green") as s:
            for item in enumerate(items):
                s.update()
    """

    _default_unicode_chars = "◓◑◒◐"
    _default_ascii_chars = "-/|\\"

    def __init__(self, msg, color="default", file=None, step=1, chars=None):
        """
        Parameters
        ----------
        msg : str
            The message to print

        color : str, optional
            An ANSI terminal color name.  Must be one of: black, red,
            green, brown, blue, magenta, cyan, lightgrey, default,
            darkgrey, lightred, lightgreen, yellow, lightblue,
            lightmagenta, lightcyan, white.

        file : :term:`file-like (writeable)`, optional
            The file to write the spinner to.  Defaults to
            `sys.stdout`.  If ``file`` is not a tty (as determined by
            calling its `isatty` member, if any, or special case hacks
            to detect the IPython console), the spinner will be
            completely silent.

        step : int, optional
            Only update the spinner every *step* steps

        chars : str, optional
            The character sequence to use for the spinner
        """
        if file is None:
            file = sys.stdout

        self._msg = msg
        self._color = color
        self._file = file
        self._step = step
        if chars is None:
            if conf.unicode_output:
                chars = self._default_unicode_chars
            else:
                chars = self._default_ascii_chars
        self._chars = chars

        self._silent = not isatty(file)

        if self._silent:
            self._iter = self._silent_iterator()
        else:
            self._iter = self._iterator()

    def _iterator(self):
        chars = self._chars
        index = 0
        file = self._file
        write = file.write
        flush = file.flush
        try_fallback = True
        terminal_width = get_terminal_size().columns
        if len(self._msg) > terminal_width:
            message = self._msg[: terminal_width - 8] + " ..."
        else:
            message = self._msg

        while True:
            write("\r")
            color_print(message, self._color, file=file, end="")
            write(" ")
            try:
                if try_fallback:
                    write = _write_with_fallback(chars[index], write, file)
                else:
                    write(chars[index])
            except UnicodeError:
                # If even _write_with_fallback failed for any reason just give
                # up on trying to use the unicode characters
                chars = self._default_ascii_chars
                write(chars[index])
                try_fallback = False  # No good will come of using this again
            flush()
            yield

            for _ in range(self._step):
                yield

            index = (index + 1) % len(chars)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        file = self._file
        write = file.write
        flush = file.flush

        if not self._silent:
            write("\r")
            color_print(self._msg, self._color, file=file, end="")
        if exc_type is None:
            color_print(" [Done]", "green", file=file)
        else:
            color_print(" [Failed]", "red", file=file)
        flush()

    def __iter__(self):
        return self

    def __next__(self):
        next(self._iter)

    def update(self, value=None):
        """Update the spin wheel in the terminal.

        Parameters
        ----------
        value : int, optional
            Ignored (present just for compatibility with `ProgressBar.update`).

        """
        next(self)

    def _silent_iterator(self):
        color_print(self._msg, self._color, file=self._file, end="")
        self._file.flush()

        while True:
            yield


class ProgressBarOrSpinner:
    """
    A class that displays either a `ProgressBar` or `Spinner`
    depending on whether the total size of the operation is
    known or not.

    It is designed to be used with the ``with`` statement::

        if file.has_length():
            length = file.get_length()
        else:
            length = None
        bytes_read = 0
        with ProgressBarOrSpinner(length) as bar:
            while file.read(blocksize):
                bytes_read += blocksize
                bar.update(bytes_read)
    """

    def __init__(self, total, msg, color="default", file=None):
        """
        Parameters
        ----------
        total : int or None
            If an int, the number of increments in the process being
            tracked and a `ProgressBar` is displayed.  If `None`, a
            `Spinner` is displayed.

        msg : str
            The message to display above the `ProgressBar` or
            alongside the `Spinner`.

        color : str, optional
            The color of ``msg``, if any.  Must be an ANSI terminal
            color name.  Must be one of: black, red, green, brown,
            blue, magenta, cyan, lightgrey, default, darkgrey,
            lightred, lightgreen, yellow, lightblue, lightmagenta,
            lightcyan, white.

        file : :term:`file-like (writeable)`, optional
            The file to write the to.  Defaults to `sys.stdout`.  If
            ``file`` is not a tty (as determined by calling its `isatty`
            member, if any), only ``msg`` will be displayed: the
            `ProgressBar` or `Spinner` will be silent.
        """
        if file is None:
            file = sys.stdout

        if total is None or not isatty(file):
            self._is_spinner = True
            self._obj = Spinner(msg, color=color, file=file)
        else:
            self._is_spinner = False
            color_print(msg, color, file=file)
            self._obj = ProgressBar(total, file=file)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return self._obj.__exit__(exc_type, exc_value, traceback)

    def update(self, value):
        """
        Update the progress bar to the given value (out of the total
        given to the constructor.
        """
        self._obj.update(value)


def print_code_line(line, col=None, file=None, tabwidth=8, width=70):
    """
    Prints a line of source code, highlighting a particular character
    position in the line.  Useful for displaying the context of error
    messages.

    If the line is more than ``width`` characters, the line is truncated
    accordingly and '…' characters are inserted at the front and/or
    end.

    It looks like this::

        there_is_a_syntax_error_here :
                                     ^

    Parameters
    ----------
    line : unicode
        The line of code to display

    col : int, optional
        The character in the line to highlight.  ``col`` must be less
        than ``len(line)``.

    file : :term:`file-like (writeable)`, optional
        Where to write to.  Defaults to `sys.stdout`.

    tabwidth : int, optional
        The number of spaces per tab (``'\\t'``) character.  Default
        is 8.  All tabs will be converted to spaces to ensure that the
        caret lines up with the correct column.

    width : int, optional
        The width of the display, beyond which the line will be
        truncated.  Defaults to 70 (this matches the default in the
        standard library's `textwrap` module).
    """
    if file is None:
        file = sys.stdout

    if conf.unicode_output:
        ellipsis = "…"
    else:
        ellipsis = "..."

    write = file.write

    if col is not None:
        if col >= len(line):
            raise ValueError("col must be less than the line length.")
        ntabs = line[:col].count("\t")
        col += ntabs * (tabwidth - 1)

    line = line.rstrip("\n")
    line = line.replace("\t", " " * tabwidth)

    if col is not None and col > width:
        new_col = min(width // 2, len(line) - col)
        offset = col - new_col
        line = line[offset + len(ellipsis) :]
        width -= len(ellipsis)
        new_col = col
        col -= offset
        color_print(ellipsis, "darkgrey", file=file, end="")

    if len(line) > width:
        write(line[: width - len(ellipsis)])
        color_print(ellipsis, "darkgrey", file=file)
    else:
        write(line)
        write("\n")

    if col is not None:
        write(" " * col)
        color_print("^", "red", file=file)


# The following three Getch* classes implement unbuffered character reading from
# stdin on Windows, and Unix.  This is taken directly from ActiveState
# Code Recipes:
# http://code.activestate.com/recipes/134892-getch-like-unbuffered-character-reading-from-stdin/
#


class Getch:
    """Get a single character from standard input without screen echo.

    Returns
    -------
    char : str (one character)
    """

    def __init__(self):
        try:
            self.impl = _GetchWindows()
        except ImportError:
            self.impl = _GetchUnix()

    def __call__(self):
        return self.impl()


class _GetchUnix:
    def __init__(self):
        import sys  # noqa: F401
        import tty  # noqa: F401

    def __call__(self):
        import sys
        import termios
        import tty

        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch


class _GetchWindows:
    def __init__(self):
        import msvcrt  # noqa: F401

    def __call__(self):
        import msvcrt

        return msvcrt.getch()
